""" pyglet_helper.util.quadric contains objects required for drawing
quadric-type
geometric shapes
"""
try:
    import pyglet.gl as gl
except ImportError:
    gl = None
from enum import Enum


class DrawingStyle(Enum):
    """
    An enumeration of the possible quadric drawing styles.
    """
    POINT = 1
    LINE = 2
    FILL = 3
    SILHOUETTE = 4


class NormalStyle(Enum):
    """
    An enumeration of the possible options when generating a quadric's normals.
    """
    NONE = 1
    FLAT = 2
    SMOOTH = 3


class Orientation(Enum):
    """
    An enumeration of the possible winding styles when rendering a quadric.
    """
    OUTSIDE = 1
    INSIDE = 2


class Quadric(object):
    """
    Generates a Quadric. Used for cylinders, spheres, and disks.
    """
    def __init__(self):
        self.quadric = gl.glu.gluNewQuadric()
        gl.glu.gluQuadricDrawStyle(self.quadric, gl.glu.GLU_FILL)
        gl.glu.gluQuadricNormals(self.quadric, gl.glu.GLU_SMOOTH)
        gl.glu.gluQuadricOrientation(self.quadric, gl.glu.GLU_OUTSIDE)
        self._drawing_style = 1
        self._normal_style = 1
        self._orientation = 1

    def __del__(self):
        gl.glu.gluDeleteQuadric(self.quadric)

    @property
    def drawing_style(self):
        """
        Get the quadric's drawing style
        :return:
        """
        return self._drawing_style

    @drawing_style.setter
    def drawing_style(self, style):
        """ Set the quadric's drawing style.

        :param style: the quadric's drawing style
        :type style: int
        """
        self._drawing_style = style
        if style == DrawingStyle.POINT:
            gl.glu.gluQuadricDrawStyle(self.quadric, gl.glu.GLU_POINT)
        elif style == DrawingStyle.LINE:
            gl.glu.gluQuadricDrawStyle(self.quadric, gl.glu.GLU_LINE)
        elif style == DrawingStyle.FILL:
            gl.glu.gluQuadricDrawStyle(self.quadric, gl.glu.GLU_FILL)
        elif style == DrawingStyle.SILHOUETTE:
            gl.glu.gluQuadricDrawStyle(self.quadric, gl.glu.GLU_SILHOUETTE)

    @property
    def normal_style(self):
        """
        Get the quadric's normal generation style
        :return:
        """
        return self._normal_style

    @normal_style.setter
    def normal_style(self, style):
        """ Set the quadric's normal generation style.

        :param style: The quadric's normal generation style
        :type style: int
        """
        self._normal_style = style
        if style == NormalStyle.NONE:
            gl.glu.gluQuadricNormals(self.quadric, gl.glu.GLU_NONE)
        elif style == NormalStyle.FLAT:
            gl.glu.gluQuadricNormals(self.quadric, gl.glu.GLU_FLAT)
        elif style == NormalStyle.SMOOTH:
            gl.glu.gluQuadricNormals(self.quadric, gl.glu.GLU_SMOOTH)

    @property
    def orientation(self):
        """
        Get the quadric's current orientation, for rendering
        :return:
        """
        return self._orientation

    @orientation.setter
    def orientation(self, side):
        """ Set the quadric's orientation direction, for rendering

        :param side: The winding orientation
        :type side: int
        """
        self._orientation = side
        if side == Orientation.OUTSIDE:
            gl.glu.gluQuadricOrientation(self.quadric, gl.glu.GLU_OUTSIDE)
        else:
            gl.glu.gluQuadricOrientation(self.quadric, gl.glu.GLU_INSIDE)

    def render_sphere(self, radius, slices, stacks):
        """ Render a sphere.

        :param radius: The radius of the sphere
        :type radius: float
        :param slices: The number of longitudinal lines
        :type slices: int
        :param stacks: The number of latitudinal lines
        :type stacks: int
        """
        gl.glu.gluSphere(self.quadric, radius, slices, stacks)

    def render_cylinder(self, base_radius, height, slices, stacks,
                        top_radius=None):
        """ Generate the polygons for a cylinder

        :param base_radius: The radius of the bottom of the cylinder.
        :type base_radius: float
        :param height: The cylinder's height
        :type height: float
        :param slices: The number of longitudinal lines
        :type slices: int
        :param stacks: The number of latitudinal lines
        :type stacks: int
        :param top_radius: The radius of the top of the cylinder. If undefined,
         the top radius will be the same as the base radius
        :type top_radius: float
        """
        # rotate the cylinder so that it is drawn along the VPython axis
        # convention
        gl.glRotatef(90, 0, 1, 0)
        if top_radius is None:
            gl.glu.gluCylinder(self.quadric, base_radius, base_radius,
                               height, slices, stacks)
        else:
            gl.glu.gluCylinder(self.quadric, base_radius, top_radius,
                               height, slices, stacks)
        gl.glRotatef(-90, 0, 1, 0)

    def render_disk(self, radius, slices, rings, rotation):
        """ Generate the polygons for a disk

        :param radius: The disk's radius
        :type radius: float
        :param slices: The number of longitudinal lines
        :type slices: float
        :param rings: The number of concentric rings in the disk, for rendering
        :type rings: int
        :param rotation: The rotation of the disk along the z-axis
        :type rotation: float
        """
        # rotate the disk so that it is drawn along the VPython axis convention
        gl.glRotatef(90, 0, gl.GLfloat(rotation), 0)
        gl.glu.gluDisk(self.quadric, 0.0, radius, slices, rings)
        gl.glRotatef(-90, 0, gl.GLfloat(rotation), 0)



"""
pyglet_helper.cone contains an object for drawing a cone
"""
try:
   import pyglet.gl as gl
except Exception as error_msg:
    gl = None
from pyglet_helper.objects import Axial
from pyglet_helper.util import Quadric, Rgb, Vector


class Cone(Axial):
    """
    A Cone object
    """
    def __init__(self, radius=1.0, color=Rgb(), pos=Vector([0, 0, 0]),
                 axis=Vector([1, 0, 0])):
        """

        :param radius: The cone's bottom radius.
        :type radius: float
        :param color: The object's color.
        :type color: pyglet_helper.util.Rgb
        :param pos: The object's position.
        :type pos: pyglet_helper.util.Vector
        :param axis: The cone points from the base to the point along the axis.
        :type axis: pyglet_helper.util.Vector
        :return:
        """
        super(Cone, self).__init__(radius=radius, color=color, pos=Vector(pos))
        self.axis = Vector(axis)

    def init_model(self, scene):
        """Add the cone quadrics to the view.

        :param scene: The view to render the model to.
        :type scene: pyglet_helper.objects.View
        """
        # The number of faces corresponding to each level of detail.
        n_sides = [8, 16, 32, 46, 68, 90]
        n_stacks = [1, 2, 4, 7, 10, 14]
        for i in range(0, 6):
            scene.cone_model[i].gl_compile_begin()
            _quadric = Quadric()
            _quadric.render_cylinder(1.0, 1.0, n_sides[i], n_stacks[i],
                                     top_radius=0.0)
            _quadric.render_disk(1.0, n_sides[i], n_stacks[i] * 2, -1)
            scene.cone_model[i].gl_compile_end()

    def render(self, scene):
        """Add the cone to the scene.

        :param scene: The view to render the model into
        :type scene: pyglet_helper.objects.View
        """
        if self.radius == 0:
            return

        self.init_model(scene)

        coverage_levels = [10, 30, 90, 250, 450]
        lod = self.lod_adjust(scene, coverage_levels, self.pos, self.radius)

        length = self.axis.mag()
        gl.glPushMatrix()
        self.model_world_transform(scene.gcf, Vector([length, self.radius,
                                                      self.radius])).gl_mult()

        self.color.gl_set(self.opacity)
        if self.translucent:
            gl.glEnable(gl.GL_CULL_FACE)

            # Render the back half.
            gl.glCullFace(gl.GL_FRONT)
            scene.cone_model[lod].gl_render()

            # Render the front half.
            gl.glCullFace(gl.GL_BACK)
            scene.cone_model[lod].gl_render()
        else:
            scene.cone_model[lod].gl_render()
        gl.glPopMatrix()

    @property
    def center(self):
        """ Get the center of the object.

        :return: position + axis/2
        :rtype: float
        """
        return self.pos + self.axis / 2.0

    @property
    def degenerate(self):
        """

        :return: True if radius == 0 or the length is 0
        :rtype: bool
        """
        return self.radius == 0.0 or self.axis.mag() == 0.0



"""
pyglet_helper.ellipsoid contains an object for drawing an ellipsoid
"""
from pyglet_helper.objects import Sphere
from pyglet_helper.util import Rgb, Vector


class Ellipsoid(Sphere):
    """
    An Ellipsoid object
    """
    def __init__(self, height=1.0, width=1.0, length=1.0, color=Rgb(),
                 pos=Vector([0, 0, 0]), axis=Vector([1.0, 0.0, 0.0])):
        """

        :param width: The ellipsoid's width.
        :type width: float
        :param height: The ellipsoid's height.
        :type height: float
        :param length: The ellipsoid's length.
        :type length: float
        :param color: The object's color.
        :type color: pyglet_helper.util.Rgb
        :param pos: The object's position.
        :type pos: pyglet_helper.util.Vector
        """
        super(Ellipsoid, self).__init__(color=color, pos=pos, axis=axis)
        self._height = None
        self._width = None
        self.height = height
        self.width = width
        self.length = length


    @property
    def scale(self):
        """
        Gets the scaling factor of the ellipsoid (1/2 of the size)
        :return: the scaling factors
        :rtype: pyglet_helper.util.Vector
        """
        return Vector([self.axis.mag(), self.height, self.width]) * 0.5

    @property
    def degenerate(self):
        """
        True if the scale in any dimension is zero, or if the object is not
        visible.
        :return: whether the object is degenerate
        :rtype: bool
        """
        return not self.visible or self.height == 0.0 or self.width == 0.0 or \
               self.axis.mag() == 0.0



"""pyglet_helper.axial contains a base class for objects with radial symmetry"""
from pyglet_helper.objects import Primitive
from math import pi
from pyglet_helper.util import Rgb, Tmatrix, Vector
from pyglet_helper.objects.material import Material


class Axial(Primitive):
    """
    A subclass for all shapes with some radial symmetry around an axis (spheres,
     cones, etc., ).
    """
    def __init__(self, axis=Vector([1, 0, 0]), radius=1.0, color=Rgb(),
                 pos=Vector([0, 0, 0]), material=Material(), other=None):
        """

        :param other: another axial object to copy properties from (optional)
        :type other: pyglet_helper.objects.Axial
        :param axis: The axis for the orientation of the object.
        :type axis: pyglet_helper.util.Vector
        :param radius: The object's radius.
        :type radius: float
        :param color: The object's color.
        :type color: pyglet_helper.util.Rgb
        :param pos: The object's position.
        :type pos: pyglet_helper.util.Vector
        :param material: The object's material
        :type material: pyglet_helper.util.Material
        """
        super(Axial, self).__init__(color=color, pos=pos, axis=axis,
                                    material=material)
        self._radius = None
        if other is not None:
            self.radius = other.radius
        else:
            self.radius = radius

    @property
    def scale(self):
        """
        Get the current scale - in the case of an axial object, this is the
        length of the axis vector
        :return: self.axis
        :rtype: pyglet_helper.util.Vector
        """
        return self.axis

    @property
    def radius(self):
        """
        Get the object's radius
        :return: the object's radius
        :rtype: float
        """
        return self._radius

    @radius.setter
    def radius(self, new_radius):
        """
        Set the object's radius
        :param new_radius: the new radius
        :return:
        """
        self._radius = new_radius

    @property
    def material_matrix(self):
        """
        Creates a transformation matrix scaled to the size of the axial object
        :return: the transformation matrix
        :return: pyglet_helper.util.Tmatrix
        """
        out = Tmatrix()
        out.translate(Vector([.0005, .5, .5]))
        out.scale(self.scale * (.999 / max(self.scale.x_component,
                                           self.scale.y_component * 2)))
        out_vector = Vector([0, 1, 0])
        out_vector = out_vector.rotate(angle=.5*pi)
        out = out * out_vector
        return out



"""
pyglet_helper.arrow contains an object for drawing an arrow
"""
try:
    from pyglet.gl import gl
except Exception as error_msg:
    gl = None
from pyglet_helper.objects import Box, Material, Primitive, Pyramid
from pyglet_helper.util import Rgb, Tmatrix, Vector


class Arrow(Primitive):
    """
     A 3D 4-sided arrow, with adjustable head and shaft. By default, pointing
     along +x, unit length.
    """
    def __init__(self, fixed_width=False, head_width=0.0, head_length=0.0,
                 shaft_width=0.0, color=Rgb(),
                 pos=Vector([0, 0, 0]), axis=(1, 0, 0)):
        """

        :param fixed_width: if True, the arrow's head width and length will not
         be scaled in proportion to its length.
        :type fixed_width: bool
        :param head_width: The width of the arrow's head section.
        :type head_width: float
        :param head_length: The length of the arrow's length section
        :type head_length: float
        :param shaft_width: The length of the arrow's shaft section
        :type shaft_width: float
        :param color: The arrow's color.
        :type color: pyglet_helper.util.Rgb
        :param pos: The arrow's position
        :type pos: pyglet_helper.util.pos
        :param axis: The arrow's axis direction
        :type axis: pyglet_helper.util.Vector
        """
        super(Arrow, self).__init__(color=color, pos=pos, axis=axis)
        # True if the width of the point and shaft should not vary with the
        # length
        # of the arrow.
        self.fixed_width = fixed_width
        self._head_width = None
        self._head_length = None
        self._shaft_width = None
        self._fixed_width = None
        # If zero, then use automatic scaling for the width's of the parts of
        # the arrow.  If nonzero, they specify proportions for the arrow in
        # world space.
        self.head_width = head_width
        self.head_length = head_length
        self.shaft_width = shaft_width
        self.box = None
        self.pyramid = None

    @property
    def degenerate(self):
        """
        :return: True if the arrow has an effective length of 0
        :rtype: bool
        """
        return self.axis.mag() == 0.0

    @property
    def head_width(self):
        """
        Get the width of the arrow's head, if undefined, it returns the
        default value of 2x the width of the shaft, if the the shaft width
        is undefined it returns a fifth of the arrow length
        :return: the arrow's head width
        :rtype: float
        """
        if self._head_width:
            return self._head_width
        if self._shaft_width:
            return 2.0 * self._shaft_width
        return 0.2 * self.axis.mag()

    @head_width.setter
    def head_width(self, new_head_width):
        """
        Set the value of the head width
        :param new_head_width: the new value of the head width
        :type new_head_width: float
        """
        self._head_width = new_head_width

    @property
    def head_length(self):
        """
        Get the length of the arrow's head. If undefined, it returns the
        default value of 3x the shaft width. If shaft width is also
        undefined, it returns the default value of 3/10 of the arrow's length.
        :return: the arrow's head length
        :rtype: float
        """
        if self._head_length:
            return self._head_length
        if self._shaft_width:
            return 3.0 * self._shaft_width
        return 0.3 * self.axis.mag()

    @head_length.setter
    def head_length(self, new_head_length):
        """
        Set the arrow's head length
        :param new_head_length: the new head length
        :type new_head_length: float
        :return:
        """
        self._head_length = new_head_length

    @property
    def shaft_width(self):
        """
        Get the arrow's shaft width. If undefined, it returns the default
        value of a tenth of the length of the arrow
        :return:
        """
        if self._shaft_width:
            return self._shaft_width
        return 0.1 * self.axis.mag()

    @shaft_width.setter
    def shaft_width(self, new_shaft_width):
        """
        set the arrow's shaft width. Once set, the value fixed_width is set
        to true, meaning default values will not be calculated on render
        :param new_shaft_width: the new shaft width
        :type new_shaft_width: float
        :return:
        """
        self._shaft_width = new_shaft_width
        self._fixed_width = True

    @property
    def fixed_width(self):
        """
        Returns true if the shaft_width property has been set,
        false otherwise. If fixed_width is false, default values for the
        shaft width, head with and head length will be calculated on render.
        :rtype: bool
        """
        return self._fixed_width

    @fixed_width.setter
    def fixed_width(self, fixed):
        """
        Sets the fixed width boolean parameter. If fixed_width is false,
        default values for the shaft width, head with and head length will be
        calculated on render.
        :param fixed: the new value for the fixed width parameter
        :type fixed: bool
        :return:
        """
        self._fixed_width = fixed

    @property
    def center(self):
        """ Calculates the center of the arrow by subtracting the axis off of
        the position

        :return: The arrow's center (position+axis)/2
        :rtype: float
        """
        return (self.pos + self.axis) / 2.0

    def render(self, scene):
        """ Render the arrow on the current view.

        :param scene: The view to render the model into.
        :type scene: pyglet_helper.objects.View
        """
        mat = Material()
        if self.degenerate:
            return
        self.init_model(scene)
        self.color.gl_set(self.opacity)
        _head_width, _shaft_width, _len, _head_length = \
            self.effective_geometry(1.0)
        if self.mat and self.mat.get_shader_program():
            model_material_loc = self.mat.get_shader_program().\
                get_uniform_location(scene, "model_material")
        else:
            model_material_loc = -1
        # Render the shaft and the head in back to front order (the shaft is in
        # front of the head if axis points away from the camera)
        shaft = self.axis.dot(scene.camera - (self.pos + self.axis *
                                              (1 - _head_length / _len))) < 0
        gl.glPushMatrix()
        self.model_world_transform(scene.gcf).gl_mult()

        for part in range(0, 2):
            if part == shaft:
                gl.glScaled(_len - _head_length, _shaft_width, _shaft_width)
                gl.glTranslated(0.5, 0, 0)
                if model_material_loc >= 0:
                    model_mat = Tmatrix()
                    scale = 1.0 / max(_len, _head_width)
                    _translation_magnitude = (_len - _head_length) * scale *\
                                             0.5
                    model_mat.translate(Vector([_translation_magnitude, 0.5,
                                                0.5]))
                    model_mat.scale(Vector([(_len - _head_length),
                                            _shaft_width,
                                            _shaft_width]) * scale)
                    mat.get_shader_program().\
                        set_uniform_matrix(scene, model_material_loc,
                                           model_mat)
                scene.box_model.gl_render()
                gl.glTranslated(-0.5, 0, 0)
                gl.glScaled(1 / (_len - _head_length), 1 / _shaft_width,
                         1 / _shaft_width)
            else:
                gl.glTranslated(_len - _head_length, 0, 0)
                gl.glScaled(_head_length, _head_width, _head_width)
                if model_material_loc >= 0:
                    model_mat = Tmatrix()
                    _scale = 1.0 / max(_len, _head_width)
                    model_mat.translate(Vector([(_len - _head_length) * _scale,
                                                0.5, 0.5]))
                    model_mat.scale(Vector([_head_length, _head_width,
                                            _head_width]) * _scale)
                    mat.get_shader_program().\
                        set_uniform_matrix(scene, model_material_loc,
                                           model_mat)
                scene.pyramid_model.gl_render()
                gl.glScaled(1 / _head_length, 1 / _head_width, 1 / _head_width)
                gl.glTranslated(-_len + _head_length, 0, 0)
        gl.glPopMatrix()

    def init_model(self, scene):
        """Add the arrow head and shaft to the scene.

        :param scene: The view to render the model into.
        :type scene: pyglet_helper.objects.View
        """
        if not scene.box_model.compiled:
            self.box = Box()
            self.box.init_model(scene)
        if not scene.pyramid_model.compiled:
            self.pyramid = Pyramid()
            self.pyramid.init_model(scene)

    def effective_geometry(self, gcf):
        """Initializes these four variables with the effective geometry for the
        arrow.  The resulting geometry is scaled to view space, but oriented
        and positioned in model space.  The only required transforms are
        reorientation and translation.

        :param gcf: The scaling factor
        :type gcf: float
        """
        # First calculate the actual geometry based on the specs for headwidth,
        # shaftwidth, shaftlength, and fixedwidth.  This geometry is calculated
        # in world space and multiplied
        min_sw = 0.02  # minimum shaftwidth
        def_sw = 0.1  # default shaftwidth
        def_hw = 2.0  # default headwidth multiplier. (x shaftwidth)
        def_hl = 3.0  # default headlength multiplier. (x shaftwidth)

        # maximum fraction of the total arrow length allocated to the head.
        max_head_length = 0.5

        eff_length = self.axis.mag() * gcf

        if self.shaft_width:
            eff_shaft_width = self.shaft_width * gcf
        else:
            eff_shaft_width = eff_length * def_sw

        if self.head_width:
            eff_head_width = self.head_width * gcf
        else:
            eff_head_width = eff_shaft_width * def_hw

        if self.head_length:
            eff_head_length = self.head_length * gcf
        else:
            eff_head_length = eff_shaft_width * def_hl

        if self.fixed_width:
            if eff_head_length > max_head_length * eff_length:
                eff_head_length = max_head_length * eff_length
        else:
            if eff_shaft_width < eff_length * min_sw:
                scale = eff_length * min_sw / eff_shaft_width
                eff_shaft_width = eff_length * min_sw
                eff_head_width *= scale
                eff_head_length *= scale
            if eff_head_length > eff_length * max_head_length:
                scale = eff_length * max_head_length / eff_head_length
                eff_head_length = eff_length * max_head_length
                eff_head_width *= scale
                eff_shaft_width *= scale
        return [eff_head_width, eff_shaft_width, eff_length, eff_head_length]



"""
pyglet_helper.util.rgba contains object to describe colors
"""
try:
    import pyglet.gl as gl
except Exception as error_msg:
    gl = None


class Rgba(object):
    """
    Defines a color to be used by OpenGl, including RGB and opacity.
    """
    def __init__(self, red=1.0, green=1.0, blue=1.0, opacity=1.0, color=None):
        """
        :param red: The red value of the color, value between 0 and 1
        :type red: float
        :param green: The green value of the color, value between 0 and 1
        :type green: float
        :param blue: The blue value of the color, value between 0 and 1
        :type blue: float
        :param opacity: The opacity value of the color, value between 0 and 1
        :type opacity: float
        :param c: A list of values to copy into a new color
        :type c: list
        """
        if color is not None:
            if len(color) == 4:
                for i in range(4):
                    self[i] = color[i]
            else:
                raise ValueError("RGBA color vector must be of length 4!")
        else:
            self.red = red
            self.green = green
            self.blue = blue
            self.opacity = opacity

    def __getitem__(self, item):
        if item == 0:
            return self.red
        elif item == 1:
            return self.green
        elif item == 2:
            return self.blue
        elif item == 3:
            return self.opacity
        else:
            raise ValueError("no such component")

    def __setitem__(self, item, value):
        if item == 0:
            self.red = value
        elif item == 1:
            self.green = value
        elif item == 2:
            self.blue = value
        elif item == 3:
            self.opacity = value
        else:
            raise ValueError("no such component")

    def __str__(self):
        return "color: r" + str(self.red) + " g" + str(self.green) + " b" + \
               str(self.blue) + " o"+str(self.opacity)

    def desaturate(self):
        """ Return a desaturated version of the color

        :rtype: pyglet_helper.util.Rgba
        :return: the desaturized color.
        """
        ret = Rgb(red=self.red, green=self.green, blue=self.blue).desaturate()
        return Rgba(red=ret.red, green=ret.green, blue=ret.blue,
                    opacity=self.opacity)

    def grayscale(self):
        """ Return a version of the color projected onto a grayscale space

        :return: the color, projected onto grayscale.
        :rtype: pyglet_helper.util.Rgba()
        """
        ret = Rgb(red=self.red, green=self.green, blue=self.blue).grayscale()
        return Rgba(red=ret.red, green=ret.green, blue=ret.blue,
                    opacity=self.opacity)

    def gl_set(self):
        """
        Set this color to the current material in OpenGL.
        """
        color = (gl.GLfloat * 4)(*[self.red, self.green, self.blue,
                                   self.opacity])
        gl.glMaterialfv(gl.GL_FRONT_AND_BACK, gl.GL_AMBIENT_AND_DIFFUSE, color)
        gl.glMaterialfv(gl.GL_FRONT_AND_BACK, gl.GL_SPECULAR, color)
        gl.glMaterialf(gl.GL_FRONT_AND_BACK, gl.GL_SHININESS, 50)


class Rgb(object):
    """
    Define an RGB color to be used by OpenGl
    """
    def __init__(self, red=1.0, green=1.0, blue=1.0, color=None):
        """
        :param red: The red value of the color, value between 0 and 1
        :type red: float
        :param green: The green value of the color, value between 0 and 1
        :type green: float
        :param blue: The blue value of the color, value between 0 and 1
        :type blue: float
        :param color: A list of values to copy into a new color
        :type color: list
        """
        if color is not None:
            if len(color) == 3:
                for i in range(3):
                    self[i] = color[i]
            else:
                raise ValueError("RGB color vector must be of length 4!")
        else:
            self.red = red
            self.green = green
            self.blue = blue

    def __str__(self):
        return "color: r" + str(self.red) + " g" + str(self.green) \
               + " b" + str(self.blue)

    def __getitem__(self, item):
        if item == 0:
            return self.red
        elif item == 1:
            return self.green
        elif item == 2:
            return self.blue
        else:
            raise ValueError("no such component")

    def __setitem__(self, item, value):
        if item == 0:
            self.red = value
        elif item == 1:
            self.green = value
        elif item == 2:
            self.blue = value
        else:
            raise ValueError("no such component")

    @property
    def rgb(self):
        """
        Get a tuple with the three components of the rgba color
        :return: (red, green, blue)
        :rtype: tuple
        """
        return (self.red, self.green, self.blue)

    @rgb.setter
    def rgb(self, grayscale):
        """
        Set the current rgb color to a grayscale value
        :param grayscale: the grayscale value'
        :type grayscale: float
        """
        self.red = grayscale
        self.green = grayscale
        self.blue = grayscale

    def desaturate(self):
        """ Convert the color to HSV space, reduce the saturation by a factor
        of 0.5, then convert back to RGB space

        :rtype: pyglet_helper.util.Rgb()
        :return: the desaturated color.
        """
        cut_saturation = 0.5  # cut the saturation by this factor

        # r,g,b values are from 0 to 1
        # h = [0,360], s = [0,1], v = [0,1]
        # if s == 0, then arbitrarily set h = 0

        cmin = self.red
        if self.green < cmin:
            cmin = self.green
        if self.blue < cmin:
            cmin = self.blue

        cmax = self.red
        if self.green > cmax:
            cmax = self.green
        if self.blue > cmax:
            cmax = self.blue
        value = cmax  # v
        delta = cmax - cmin

        if cmin == cmax:  # completely unsaturated color is some gray
            # if r = g = b = 0, s = 0, v in principle undefined but set to 0
            saturation = 0.0
            hue = 0.0
        else:
            saturation = delta / cmax  # s
            if self.red == cmax:
                # between yellow & magenta
                hue = (self.green - self.blue) / delta
            elif self.green == cmax:
                # between cyan & yellow
                hue = 2.0 + (self.blue - self.red) / delta
            else:
                # between magenta & cyan
                hue = 4.0 + (self.red - self.green) / delta

            if hue < 0.0:
                hue += 6.0  # make it 0 <= h < 6

        # unsaturate somewhat to make sure both eyes have something to see
        saturation *= cut_saturation
        ret = Rgb()
        if saturation == 0.0:
            # achromatic (grey)
            ret.red = ret.green = ret.blue = value
        else:
            i = int(hue)  # hue represents sector 0 to 5
            fraction = hue - i  # fractional part of hue
            p_vandam = value * (1.0 - saturation)
            q_vandam = value * (1.0 - saturation * fraction)
            t_vandam = value * (1.0 - saturation * (1.0 - fraction))

            if i == 0:
                ret.red = value
                ret.green = t_vandam
                ret.blue = p_vandam
            elif i == 1:
                ret.red = q_vandam
                ret.green = value
                ret.blue = p_vandam
            elif i == 2:
                ret.red = p_vandam
                ret.green = value
                ret.blue = t_vandam
            elif i == 3:
                ret.red = p_vandam
                ret.green = q_vandam
                ret.blue = value
            elif i == 4:
                ret.red = t_vandam
                ret.green = p_vandam
                ret.blue = value
            else:  # case 5:
                ret.red = value
                ret.green = p_vandam
                ret.blue = q_vandam
        return ret

    def grayscale(self):
        """ Convert the color to grayscale

        :return: the color, projected onto grayscale.
        :rtype: pyglet_helper.util.Rgb()
        """
        # The constants 0.299, 0.587, and 0.114 are intended to account for the
        # relative intensity of each color to the human eye.
        gamma = 2.5
        black = pow(0.299 * pow(self.red, gamma) + 0.587 * pow(self.green,
                                                               gamma)
                    + 0.114 * pow(self.blue, gamma), 1.0 / gamma)
        return Rgb(red=black, green=black, blue=black)

    def gl_set(self, opacity=1.0):
        """ Set the current material to this color

        :param opacity: the opacity value of the color
        :type opacity: float
        """
        color = (gl.GLfloat * 4)(*[self.red, self.green, self.blue, opacity])
        gl.glMaterialfv(gl.GL_FRONT_AND_BACK, gl.GL_AMBIENT_AND_DIFFUSE, color)
        gl.glMaterialfv(gl.GL_FRONT_AND_BACK, gl.GL_SPECULAR, color)
        gl.glMaterialf(gl.GL_FRONT_AND_BACK, gl.GL_SHININESS, 50)



"""
The following file contains definitions for GL functions when pyglet can't
be used, such as on testing on continuous integration systems.
"""
from ctypes import c_float, c_double, c_uint


GLfloat = c_float
GLdouble = c_double
GLuint = c_uint
GL_COLOR_BUFFER_BIT = 16384
GL_CULL_FACE = 2884
GL_DEPTH_BUFFER_BIT = 256
GL_DEPTH_TEST = 2929
GL_LIGHTING = 2896
GL_LIGHT0 = 16384
GL_LIGHT1 = 16385
GL_LIGHT2 = 16386
GL_LIGHT3 = 16387
GL_LIGHT4 = 16388
GL_LIGHT5 = 16389
GL_LIGHT6 = 16390
GL_LIGHT7 = 16391


GL_COMPILE = 0
GL_FRONT = 0
GL_BACK = 0
GL_FLOAT = 0
GL_MODELVIEW_MATRIX = 0
GL_TEXTURE_MATRIX = 0
GL_COLOR_MATRIX = 0
GL_PROJECTION_MATRIX = 0
GL_FRONT_AND_BACK = 0
GL_AMBIENT_AND_DIFFUSE = 0
GL_SPECULAR = 0
GL_SHININESS = 0
GL_CLIENT_VERTEX_ARRAY_BIT = 0
GL_VERTEX_ARRAY = 0
GL_NORMAL_ARRAY = 0
GL_TRIANGLES = 0
GL_UNSIGNED_INT = 0
GL_VERTEX_SHADER_ARB = 0
GL_FRAGMENT_SHADER_ARB = 0
GL_OBJECT_LINK_STATUS_ARB = 0
GL_OBJECT_INFO_LOG_LENGTH_ARB = 0
GL_POSITION = 0
GL_DIFFUSE = 0

class GLException(Exception):
   def __init__(self, value):
        self.value = value
   def __str__(self):
        return repr(self.value)


def glNormal3f(a, b, c):
    pass


def glVertex3f(a, b, c):
    pass


def glNormal3dv(component):
    pass


def glVertex3d(a, b, c):
    pass


def glVertex4d(a, b, c, d):
    pass


def glGenLists(handle):
    return handle


def glEnable(lighting):
    pass


def glClearColor(a, b, c, d):
    pass


def glColor3f(a, b, c):
    pass


def glClear(a):
    pass


def glLoadIdentity():
    pass


def glRotatef(x, y, z, w):
    pass


def glNewList(handle, compile):
    pass


def glEndList(handle=None):
    pass


def glPushMatrix():
    pass


def glCullFace(front):
    pass


def glCallList(handle):
    pass


def glPopMatrix():
    pass


def glLoadMatrixd(ctypes_matrix):
    pass


def glMultMatrixd(ctype_matrix):
    pass


def glGetFloatv(matrix, ctypes_matrix):
    return [-1.0, -1.0, -1.0, -1.0]


def glMaterialf(front_and_back, shininess, val):
    pass


def glMaterialfv(front_and_back, shininess, val):
    pass


def glPushClientAttrib(array_bit):
    pass


def glEnableClientState(array):
    pass


def glVertexPointer(pointer, type, index, vertices):
    pass


def glNormalPointer(type, index, normals):
    pass

def glDrawElements(triangles, indices_leng, type, indices):
    pass


def glPopClientAttrib():
    pass


def glBegin(triangles):
    pass


def glNormal3fv(a, b, c):
    pass


def glEnd():
    pass


def glDisable(cull_face):
    pass


def glTranslatef(a, b, c):
    pass


def glTranslated(x, y, z):
    pass


def glScaled(x, y, z):
    pass


def glLightfv(defined_light, position, light_position):
    pass


class glext_arb(object):
    GL_ARB_shader_objects = 1

    @staticmethod
    def glGetUniformLocationARB(program, name):
        return 0

    @staticmethod
    def glCreateShaderObjectARB(shader_type):
        return 1

    @staticmethod
    def glShaderSourceARB(shader, handle):
        pass

    @staticmethod
    def glCompileShaderARB(shader):
        pass

    @staticmethod
    def glAttachObjectARB(program, shader):
        pass

    @staticmethod
    def glDeleteObjectARB(shader):
        pass

    @staticmethod
    def glCreateProgramObjectARB():
        pass

    @staticmethod
    def glLinkProgramARB(program):
        pass

    @staticmethod
    def glGetObjectParameterivARB(program, link_status):
        return False

    @staticmethod
    def glGetInfoLogARB(program, length):
        return (0.0, 0.0)

class glu(object):
    GLU_FILL = 1
    GLU_SMOOTH = 1
    GLU_OUTSIDE = 1
    GLU_INSIDE = 1
    GLU_POINT = 1
    GLU_LINE = 1
    GLU_FILL = 1
    GLU_SILHOUETTE = 1

    @staticmethod
    def gluDisk(quadric, thickness, radius, slices, rings):
        pass

    @staticmethod
    def gluCylinder(quadric, base_radius, top_radius, height, slices, stacks):
        pass

    @staticmethod
    def gluNewQuadric():
        return -1

    @staticmethod
    def gluQuadricDrawStyle(quadric, fill):
        pass

    @staticmethod
    def gluQuadricNormals(quadric, smooth):
        pass

    @staticmethod
    def gluQuadricOrientation(quadric, outside):
        pass

    @staticmethod
    def gluQuadricDrawStyle(quadric, point):
        pass

    @staticmethod
    def gluQuadricNormals(quadric, style):
        pass

    @staticmethod
    def gluSphere(quadric, radius, slices, stacks):
        pass

    @staticmethod
    def gluDeleteQuadric(quadric):
        pass




""" pyglet_helper.util.shader_program contains objects for creating programs to
describe how the light treats the object
"""
try:
    import pyglet.gl as gl
except ImportError:
    gl = None


class ShaderProgram(object):
    """
    An interface for storing and executing shader programs
    """
    def __init__(self, source=None):
        self._source = None
        self.source = source
        self.program = 0
        self.uniforms = {'': 0}

    @property
    def source(self):
        """
        Get the shader program's source as a string
        :return: the shader program's source
        :rtype: str
        """
        return self._source

    @source.setter
    def source(self, source):
        """
        Set the shader program's source
        :param source: valid shader program
        :type source: str
        """
        self._source = source

    def uniform_location(self, name):
        """
        Get the location of the shader by name in the view.
        :param view: the view to inspect
        :type view: str
        :param name: the View to examine
        :type name: pyglet_helper.objects.view
        :return:
        """
        if self.program <= 0 or not gl.glext_arb.GL_ARB_shader_objects:
            return -1
        cache = self.uniforms[name]
        if cache == 0:
            cache = 2 + gl.glext_arb.glGetUniformLocationARB(self.program, name)
        return cache - 2

    def realize(self, view):
        """
        Render the shader program
        :param view: the view to render the shader program in
        :type view: pyglet_helper.objects.View
        """
        if self.program != -1:
            return

        if not view.enable_shaders:
            return

        if not gl.glext_arb.GL_ARB_shader_objects:
            return

        self.program = gl.glext_arb.glCreateProgramObjectARB()

        self.compile(gl.GL_VERTEX_SHADER_ARB)
        self.compile(gl.GL_FRAGMENT_SHADER_ARB)

        gl.glext_arb.glLinkProgramARB(self.program)

        # Check if linking succeeded
        link_ok = gl.glext_arb.glGetObjectParameterivARB\
            (self.program, gl.GL_OBJECT_LINK_STATUS_ARB)

        if not link_ok:
            # Some drivers (incorrectly?) set the GL error in
            # glLinkProgramARB() in this situation
            length = gl.glext_arb.\
                glGetObjectParameterivARB(self.program,
                                          gl.GL_OBJECT_INFO_LOG_LENGTH_ARB)
            temp = ['a'] * (length + 2)
            length, temp[0] = gl.glext_arb.glGetInfoLogARB(self.program,
                                                         length + 1)

            print("VPython WARNING: errors in shader program:\n" + str(temp)
                  + "\n")

            # Get rid of the program, since it can't be used without generating
            # GL errors.  We set program to 0 instead of -1 so that binding it
            # will revert to the fixed function pipeline, and realize() won't
            # be called again.
            gl.glext_arb.glDeleteObjectARB(self.program)
            self.program = 0
            return

    def compile(self, shader_type):
        """
        Compiles and attaches the current shader
        :param view: the view to render the shader program to
        :type view: pyglet_helper.objects.View
        :param shader_type: the shader type, e.g., GL_VERTEX_SHADER,
        GL_GEOMETRY_SHADER
        :type shader_type: valid opengl shader type
        """
        shader = gl.glext_arb.glCreateShaderObjectARB(shader_type)
        gl.glext_arb.glShaderSourceARB(shader, 1)
        gl.glext_arb.glCompileShaderARB(shader)
        gl.glext_arb.glAttachObjectARB(self.program, shader)
        gl.glext_arb.glDeleteObjectARB(shader)

    def get(self):
        """
        gets a handle to the current program
        :return: handle to the current program
        :rtype: glShaderObject
        """
        return self.program

    def gl_free(self):
        """
        Remove the current program from memory
        """
        gl.glext_arb.glDeleteObjectARB(self.program)


class UseShaderProgram(object):
    """
    A Class to handle the initialization of ShaderPrograms
    """
    def __init__(self, view, program=None):

        self.view = view
        self.m_ok = None
        self.program = program
        self.old_program = None
        self.init()

    def __exit__(self, _type, value, traceback):
        if self.old_program < 0 or not self.view.glext.ARB_shader_objects:
            return
        gl.glUseProgramObjectARB(self.old_program)

    @property
    def invoked(self):
        """
        Has the shader program been invoked?
        :return: True if the shader program was successfully invoked
        :rtype: boolean
        """
        return self.m_ok

    def init(self):
        """
        Initialize the shader program
        :return: Nothing
        """
        self.m_ok = False
        if not self.program or not self.view.glext.ARB_shader_objects or not \
                self.view.enable_shaders:
            self.old_program = -1
            return

        self.program.realize(self.view)

        # For now, nested shader invocations aren't supported.
        # old_program = v.glext.glGetHandleARB( GL_PROGRAM_OBJECT_ARB )
        self.old_program = 0

        self.view.glext.glUseProgramObjectARB(self.program.program)
        self.m_ok = (self.program.program != 0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module containing renderable objects
"""
from __future__ import absolute_import

from pyglet_helper.objects.material import Material, UNSHADED, EMISSIVE, \
    DIFFUSE, PLASTIC, ROUGH, SHINY, CHROME, ICE, GLASS, BLAZED, SILVER, \
    WOOD, MARBLE, EARTH, BLUEMARBLE, BRICKS
from pyglet_helper.objects.renderable import Renderable, View
from pyglet_helper.objects.primitive import Primitive
from pyglet_helper.objects.rectangular import Rectangular
from pyglet_helper.objects.box import Box
from pyglet_helper.objects.pyramid import Pyramid
from pyglet_helper.objects.arrow import Arrow
from pyglet_helper.objects.axial import Axial
from pyglet_helper.objects.cone import Cone
from pyglet_helper.objects.cylinder import Cylinder
from pyglet_helper.objects.sphere import Sphere
from pyglet_helper.objects.ellipsoid import Ellipsoid
from pyglet_helper.objects.light import Light
from pyglet_helper.objects.ring import Ring



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Import the fake gl modules
"""
from pyglet_helper.test.fake_gl import GL_COLOR_BUFFER_BIT, GL_CULL_FACE, \
    GL_DEPTH_BUFFER_BIT, GL_DEPTH_TEST, GL_LIGHTING, GL_LIGHT0, GL_LIGHT1, \
    GL_LIGHT2, GL_LIGHT3, GL_LIGHT4, GL_LIGHT5, GL_LIGHT6,  GL_LIGHT7, \
    glClear, glClearColor, glColor3f, glEnable, GLfloat, glGenLists, \
    glLoadIdentity, glNormal3f, glVertex3f



"""
pyglet_helper.pyramid contains an object for drawing a pyramid
"""
try:
    import pyglet.gl as gl
except Exception as error_msg:
    gl = None
from pyglet_helper.objects import Rectangular
from pyglet_helper.util import Rgb, Tmatrix, Vector


class Pyramid(Rectangular):
    """
    A Pyramid Object
    """
    def __init__(self, axis=Vector([1, 0, 0]), pos=Vector([0, 0, 0]),
                 width=1.0, height=1.0, length=1.0, color=Rgb()):
        """
        :param color: The object's color.
        :type color: pyglet_helper.util.Rgb
        :param pos: The object's position.
        :type pos: pyglet_helper.util.Vector
        :param width: The pyramid's width.
        :type width: float
        :param height: The pyramid's height.
        :type height: float
        :param length: The pyramid's length.
        :type length: float
        :return:
        """
        super(Pyramid, self).__init__(axis=axis, pos=pos, color=color,
                                      width=width, height=height,
                                      length=length)
        self.compiled = False

    def init_model(self, scene):
        """ Add the pyramid normals and vertices to the View

        :param scene: The view to render the model to.
        :type scene: pyglet_helper.objects.View
        """
        # Note that this model is also used by arrow!
        scene.pyramid_model.gl_compile_begin()

        vertices = [[0, .5, .5],
                    [0, -.5, .5],
                    [0, -.5, -.5],
                    [0, .5, -.5],
                    [1, 0, 0]]
        triangle_indices = [0, 0, 0] * 6
        triangle_indices[0] = [3, 0, 4]  # top
        triangle_indices[1] = [1, 2, 4]  # bottom
        triangle_indices[2] = [0, 1, 4]  # front
        triangle_indices[3] = [3, 4, 2]  # back
        triangle_indices[4] = [0, 3, 2]  # left (base) 1
        triangle_indices[5] = [0, 2, 1]  # left (base) 2

        normals = [[1, 2, 0], [1, -2, 0], [1, 0, 2], [1, 0, -2], [-1, 0, 0],
                   [-1, 0, 0]]

        gl.glEnable(gl.GL_CULL_FACE)
        gl.glBegin(gl.GL_TRIANGLES)

        # Inside
        for face in range(0, 6):
            gl.glNormal3f(-normals[face][0], -normals[face][1], -normals[face][2])
            for vertex in range(0, 3):
                #print triangle_indices[face]
                #print vertices[triangle_indices[face]][2 - vertex]
                vert = [gl.GLfloat(i) for i in
                        vertices[triangle_indices[face][2 - vertex]]]
                gl.glVertex3f(*vert)

        # Outside
        for face in range(0, 6):
            gl.glNormal3fv(*[gl.GLfloat(i) for i in normals[face]])
            for vertex in range(0, 3):
                gl.glVertex3f(*[gl.GLfloat(i) for i in vertices[triangle_indices[
                                face][vertex]]])

        gl.glEnd()
        gl.glDisable(gl.GL_CULL_FACE)
        self.compiled = True

        scene.pyramid_model.gl_compile_end()

    @property
    def center(self):
        """
        Gets the object's position
        :return: the object's position
        :rtype: pyglet_helper.util.Vector
        """
        return self.pos + self.height*self.axis/3.0

    @property
    def material_matrix(self):
        """
        Creates a transformation matrix for pyramid objects
        :return: the transformation matrix
        :rtype: pyglet_helper.util.Tmatrix
        """
        out = Tmatrix()
        out.translate(Vector([0, .5, .5]))
        scale = Vector([self.axis.mag(), self.height, self.width])
        out.scale(Vector([self.scale,self.scale,self.scale]) *
                  (1.0 /max(scale.x_component, max(scale.y_component,
                                              scale.z_component))))
        return out

    def render(self, scene):
        """Add a pyramid to the view.

        :param scene: The view to render the model into
        :type scene: pyglet_helper.objects.View
        """
        if not scene.pyramid_model.compiled:
            self.init_model(scene)

        self.color.gl_set(self.opacity)
        gl.glPushMatrix()
        self.apply_transform(scene)
        scene.pyramid_model.gl_render()
        gl.glPopMatrix()



""" pyglet_helper.light contains an object for creating lights
"""
try:
    import pyglet.gl as gl
except Exception as error_msg:
    gl = None
from pyglet_helper.objects import Renderable
from pyglet_helper.util import Rgb, Vector


class Light(Renderable):
    """
    A light object
    """
    def __init__(self, color=Rgb(), specular=(.5, .5, 1, 0.5),
                 diffuse=(1, 1, 1, 1), position=(1, 0.5, 1, 0)):
        """

        :param color: The object's color.
        :type color: pyglet_helper.util.Rgb
        :param specular: The color of the specular reflections on the objects
        in the scene.
        :type specular: array_like
        :param diffuse: The color of the diffuse reflections of the objects in
        the scene.
        :type position: array_like
        :param position: The object's position.
        :type position: array_like
        """
        super(Light, self).__init__(color=color)
        self.color = None
        self.rgb = color
        self.specular = (gl.GLfloat * 4)(*specular)
        self.diffuse = (gl.GLfloat * 4)(*diffuse)
        self.position = (gl.GLfloat * 4)(*position)

    @property
    def rgb(self):
        """
        Gets the light's color
        :return: the light's color
        :rtype: pyglet_helper.util.rgb
        """
        return self.color

    @rgb.setter
    def rgb(self, new_color):
        """
        Sets the light's color
        :param new_color: the new light color
        :type new_color: pyglet_helper.util.Rgb
        :return:
        """
        self.color = new_color

    @property
    def center(self):
        """
        Gets the light's center. Since the light is not a physical object,
        this will always be an empty vector
        :return: the light's center
        :rtype: pyglet_helper.util.Vector
        """
        return Vector()

    @property
    def material(self):
        """
        Though the Renderable object that Light inherits from has a
        material, the Light object does not. Will raise an exception if set
        or gotten.
        :return:
        """
        raise ValueError("light object does not have a material.")

    @material.setter
    def material(self, material):
        """
        Though the Renderable object that Light inherits from has a
        material, the Light object does not. Will raise an exception if set
        or gotten.
        :return:
        """
        _ = material
        raise ValueError("light object does not have a material.")

    @property
    def is_light(self):
        """
        Returns true if the parent object is a light
        :return: bool
        """
        return True

    def render(self, scene):
        """ Add the light to the scene.
        :param scene: The view to render the model into
        :type scene: pyglet_helper.objects.View
        """
        scene.lights.append(self)
        scene.draw_lights()




"""
VPython material tools and definitions
Stable Interfaces:
    materials.unshaded
    materials.diffuse
    materials.plastic
    materials.rough
    materials.wood
    materials.marble
    materials.earth
        Individual named materials.  The exact look of these might change in
        future versions of Visual, but the names should continue to work.
    materials.texture( data, channels=None, mipmap=True,
                       interpolate=True, name="texture" )
        Create a material by "extruding"/projecting a 2D image.
        See VPython documentation for more details.
    materials.load_tga(file)
Unstable interfaces:
    materials.shader( name, shader, version, textures=(), translucent=False )
        This is the low level interface for constructing a material based on
        a GLSL shader program.

        It is not fully documented because it is subject to change in future
        versions.  In particular, it is likely that some or all programs using
        materials.shader() directly will require modification to run in future
        versions of Visual.
        If you come up with really nice new materials, be sure to post them
        to the VPython forum, so that they can be considered for
        incorporation in new versions of Visual.  As well as benefitting other
        users, this will remove from you the burden of updating your shader for
        new versions.
        If you want to learn how to write shaders, look at some of the examples
        below (such as "wood"), and consult the GLSL reference specification
        which is available online.
        To minimize the likelihood that you will have to rewrite your shaders
        for future architectural changes in Visual, avoid accessing the
        built in gl_* variables defined in that specification; instead use
        only normal, position, mat_pos, object_color, and object_opacity as
        inputs and write your output shaded material color to material_color
        and material_opacity.  Many of these are currently simple macros
        expanding to gl_* variables, but in the future their definitions may
        change.  Use lightAt() instead of the light_* uniforms if possible.
    materials.raw_texture( data, interpolate, mipmap )
    materials.tx_turb3
    materials.tx_wood
        raw 2D textures useful only for filling in the textures parameter of
        materials.shader().
"""

from numpy import array, reshape, fromstring, ubyte, asarray
import os.path
import sys

from pyglet_helper.util.texture import Texture
from pyglet_helper.objects.material import Material, UNSHADED, EMISSIVE, \
    DIFFUSE, PLASTIC, ROUGH, SHINY, CHROME, ICE, GLASS, BLAZED, SILVER, \
    WOOD, MARBLE, EARTH, BLUEMARBLE, BRICKS


class RawTexture(Texture):
    """
    An extension of the pyglet_helper.util.Texture object to allow for
    arbitrary input data
    """
    def __init__(self, **kwargs):
        Texture.__init__(self)
        for key, value in kwargs.items():
            if key == 'data' and value is None:
                raise ValueError("Cannot nullify a texture by assigning its "
                                   "data to None")
            else:
                self.__setattr__(key, value)


class ShaderMaterial(Material):
    """
    An extension of the pyglet_helper.object.Material object to allow for
    arbitrary input data
    """
    def __init__(self, **kwargs):
        Material.__init__(self)
        for key, value in kwargs.items():
            self.__setattr__(key, value)
        if not hasattr(self, "name"):
            self.name = "no_name"


def convert_data(data):
    """
    Convert an input data object into a numpy array of unsigned bytes
    :param data:
    :return:
    """
    data = asarray(data)
    if data.dtype != ubyte:
        data = array(255 * data, ubyte)
    if len(data.shape) == 2:
        data = reshape(data, data.shape + (1,))
    return data


def load_tga(fileid):
    """
    Load a TGA image file to use as a texture
    :param fileid: the filename of image
    :type fileid: str
    :return: image data
    """
    header_length = 18 # length of header in targa file
    if isinstance(fileid, str):
        if fileid[-4:] != ".tga":
            fileid += ".tga"
        fileid = open(fileid, "rb")
    data = fromstring(fileid.read(), ubyte)
    width = data[12] + 256 * data[13]
    height = data[14] + 256 * data[15]
    _bytes = data[16] >> 3
    image = data[header_length:header_length + width * height * _bytes]
    if 1 <= _bytes <= 2:
        image = image.reshape((height, width, _bytes))
    elif 3 <= _bytes <= 4:
        # make copy; must reverse byte order bgr -> rgb
        red = image[0::_bytes].copy()
        image[0::_bytes] = image[2::_bytes]  # blue
        image[2::_bytes] = red
        image = image.reshape((height, width, _bytes))
    else:
        raise IOError("%s is not a valid targa file." % fileid)
    # Photoshop "save as targa" starts the data in lower left;
    # last byte in header is zero.
    # Visual and POV-Ray start data in upper left; last header byte is nonzero.
    if data[header_length - 1] == 0:
        image = image[::-1]
    return image

# The following code addresses a problem for those packaging a program using
# py2exe, as reported by Jason Morgan.

TEXTURE_PATH = ""

if hasattr(sys, 'frozen'):
    if getattr(sys, 'frozen') == "windows_exe" or getattr(sys, 'frozen') == \
            "console_exe":
        TEXTURE_PATH = "visual\\"
else:
    TEXTURE_PATH = os.path.split(__file__)[0] + "/"

TURBULENCE3_DATA = load_tga(str(TEXTURE_PATH) + "turbulence3")  # the targa
# file is 512*512*3
TX_TURB3 = RawTexture(data=reshape(TURBULENCE3_DATA, (64, 64, 64, 3)),
                      interpolate=True,
                      mipmap=False)
TX_WOOD = RawTexture(data=load_tga(TEXTURE_PATH + "wood"), interpolate=True)
TX_BRICK = RawTexture(data=load_tga(TEXTURE_PATH + "brickbump"),
                      interpolate=True)
DATA_R = load_tga(TEXTURE_PATH + "random")
TX_RANDOM = RawTexture(data=reshape(DATA_R, (64, 64, 64, 3)), interpolate=True,
                       mipmap=False)

LIBRARY = open(os.path.join(os.path.dirname(__file__), "library.txt"),
               "r").read()


def shader(name, _shader, version, library=LIBRARY, **kwargs):
    """
    Create a shader material by reading in from the library
    :param name: the name of the shader program in the library to select
    :type name: str
    :param _shader: the existing shader to read from, can be empty
    :type _shader: str
    :param version:
    :param library: the library to draw from
    :param kwargs:
    :return:
    """

    if isinstance(version, tuple):
        min_version, max_version = version
    else:
        min_version, max_version = version, version
    if max_version < 5.00 or min_version >= 5.10:
        raise ValueError("shader version " + str(version) + " not supported.")
    if _shader.find("[vertex]") < 0 and library:
        _shader += """
            [vertex]
            void main() {
                basic();
            }"""
    _shader = library + "\n".join([l.strip() for l in _shader.split("\n")])
    return ShaderMaterial(name=name, shader=_shader, **kwargs)


MATERIALS = [UNSHADED, EMISSIVE, DIFFUSE, PLASTIC, ROUGH, SHINY, CHROME, ICE,
             GLASS, BLAZED, SILVER, WOOD, MARBLE, EARTH, BLUEMARBLE, BRICKS]


"""
pyglet_helper.box contains an object for drawing a box to the screen
"""
try:
    import pyglet.gl as gl
except ImportError:
    gl = None
from pyglet_helper.objects import Rectangular
from pyglet_helper.util import Rgb, Vector


class Box(Rectangular):
    """
     A box object.
    """
    def __init__(self, width=1.0, height=1.0, length=1.0, color=Rgb(),
                 pos=Vector([0, 0, 0])):
        """

        :param width: The box's width.
        :type width: float
        :param height: The box's height.
        :type height: float
        :param length: The box's length.
        :type length: float
        :param color: The object's color.
        :type color: pyglet_helper.util.Rgb
        :param pos: The object's position.
        :type pos: pyglet_helper.util.Vector
        """
        super(Box, self).__init__(width=width, height=height, color=color,
                                  length=length, pos=pos)
        self.initialized = False
        self.skip_right_face = False

    def init_model(self, scene):
        """ Add the Vertexes and Normals to the compile list.

        :param scene: The view to render the model to.
        :type scene: pyglet_helper.objects.View
        """
        # Note that this model is also used by arrow!
        scene.box_model.gl_compile_begin()
        gl.glEnable(gl.GL_CULL_FACE)
        gl.glBegin(gl.GL_TRIANGLES)
        self.generate_model()
        gl.glEnd()
        gl.glDisable(gl.GL_CULL_FACE)
        scene.box_model.gl_compile_end()
        self.initialized = True

    def render(self, scene):
        """Render the box in the view

        :param scene: The view to render the model into
        :type scene: pyglet_helper.objects.View
        """
        if not scene.box_model.compiled:
            self.init_model(scene)
        self.color.gl_set(self.opacity)
        gl.glPushMatrix()
        self.apply_transform(scene)
        scene.box_model.gl_render()
        gl.glPopMatrix()

    def generate_model(self):
        """ Generate the vertices and normals.
        """
        corner = 0.5
        vertices = [
            [[+corner, +corner, +corner], [+corner, -corner, +corner],
             [+corner, -corner, -corner], [+corner, +corner, -corner]],
            # Right face
            [[-corner, +corner, -corner], [-corner, -corner, -corner],
             [-corner, -corner, +corner], [-corner, +corner, +corner]],
            # Left face
            [[-corner, -corner, +corner], [-corner, -corner, -corner],
             [+corner, -corner, -corner], [+corner, -corner, +corner]],
            # Bottom face
            [[-corner, +corner, -corner], [-corner, +corner, +corner],
             [+corner, +corner, +corner], [+corner, +corner, -corner]],
            # Top face
            [[+corner, +corner, +corner], [-corner, +corner, +corner],
             [-corner, -corner, +corner], [+corner, -corner, +corner]],
            # Front face
            [[-corner, -corner, -corner], [-corner, +corner, -corner],
             [+corner, +corner, -corner], [+corner, -corner, -corner]]
            # Back face
        ]
        normals = [[+1, 0, 0], [-1, 0, 0], [0, -1, 0],
                   [0, +1, 0], [0, 0, +1], [0, 0, -1]]
        # Draw inside (reverse winding and normals)
        for face in range(self.skip_right_face, 6):
            gl.glNormal3f(-normals[face][0], -normals[face][1],
                                 -normals[face][2])
            for vertex in range(0, 3):
                gl.glVertex3f(gl.GLfloat(vertices[face][3 - vertex][0]),
                           gl.GLfloat(vertices[face][3 - vertex][1]),
                           gl.GLfloat(vertices[face][3 - vertex][2]))
            for vertex in (0, 2, 3):
                gl.glVertex3f(gl.GLfloat(vertices[face][3 - vertex][0]),
                           gl.GLfloat(vertices[face][3 - vertex][1]),
                           gl.GLfloat(vertices[face][3 - vertex][2]))
        # Draw outside
        for face in range(self.skip_right_face, 6):
            gl.glNormal3f(gl.GLfloat(normals[face][0]),
                                 gl.GLfloat(normals[face][1]),
                       gl.GLfloat(normals[face][2]))
            for vertex in range(0, 3):
                gl.glVertex3f(gl.GLfloat(vertices[face][vertex][0]),
                           gl.GLfloat(vertices[face][vertex][1]),
                           gl.GLfloat(vertices[face][vertex][2]))
            for vertex in (0, 2, 3):
                gl.glVertex3f(gl.GLfloat(vertices[face][vertex][0]),
                           gl.GLfloat(vertices[face][vertex][1]),
                           gl.GLfloat(vertices[face][vertex][2]))


""" pyglet_helper.util.linear contains vector and vertex objects needed for
transformations and describes linear algebra operations
"""
from __future__ import division, print_function
try:
    import pyglet.gl as gl
except Exception as error_msg:
    gl = None
from numpy import matrix, identity, nditer
from numpy.linalg import inv
from math import sqrt, acos, asin, pi


class Vector(object):
    """
    A vector object used for math operations, and for storing 3D points
    """
    def __init__(self, in_vector=None):
        """
        :param in_vector: an array_like of length 3 defining the components of
        the vector
        :type in_vector: array_like
        """
        self._x_component = None
        self._y_component = None
        self._z_component = None
        if in_vector is None:
            self.x_component = 0
            self.y_component = 0
            self.z_component = 0
        if type(in_vector) is tuple or type(in_vector) is list:
            self.x_component = in_vector[0]
            self.y_component = in_vector[1]
            self.z_component = in_vector[2]
        elif hasattr(in_vector, 'x_component') and \
                hasattr(in_vector, 'y_component') and \
                hasattr(in_vector, 'z_component'):
            self.x_component = getattr(in_vector, 'x_component')
            self.y_component = getattr(in_vector, 'y_component')
            self.z_component = getattr(in_vector, 'z_component')

    def __add__(self, other):
        """
        add another vector or vertex to the current vector
        :param other: the other vector or vertex
        :type other: pyglet_helper.util.Vector or pyglet_helper.util.Vertex
        :return:
        """
        return Vector([self.x_component + other.x_component, self.y_component
                       + other.y_component, self.z_component +
                       other.z_component])

    def __sub__(self, vector):
        """
        subtract another vector or vertex to the current vector
        :param other: the other vector or vertex
        :type other: pyglet_helper.util.Vector or pyglet_helper.util.Vertex
        :return:
        """
        return Vector([self.x_component - vector.x_component, self.y_component -
                       vector.y_component,
                       self.z_component - vector.z_component])

    def __mul__(self, vector):
        """
        Multiply the vector by a float or another vector
        :param vector: the multiplication factor
        :type vector: float or Vector or Vertex
        :return: resulting vector
        :rtype: Vector
        """
        if type(vector) == type(self):
            return Vector([self.x_component * vector.x_component,
                           self.y_component * vector.y_component,
                           self.z_component * vector.z_component])
        else:
            return Vector([self.x_component * vector, self.y_component * vector,
                           self.z_component * vector])

    def __rmul__(self, vector):
        """
        Multiply the vector by a float or another vector
        :param vector: the multiplication factor
        :type vector: float or Vector or Vertex
        :return: resulting vector
        :rtype: Vector
        """
        return Vector([vector * self.x_component, vector * self.y_component,
                       vector * self.z_component])

    def __div__(self, scale):
        """
        Python 2 and 3 compatibility
         divide the current vector by a single value
        :param scale: the factor to divide the vector by
        :type scale: float
        :return: the vector divided by scale
        :rtype: pyglet_helper.util.Vector
        """
        return self.__truediv__(scale)

    def __truediv__(self, scale):
        """
        divide the current vector by a single value
        :param scale: the factor to divide the vector by
        :type scale: float
        :return: the vector divided by scale
        :rtype: pyglet_helper.util.Vector
        """
        return Vector([self.x_component / scale, self.y_component / scale,
                       self.z_component / scale])

    def __eq__(self, vector):
        """
        measures whether each component in a vector is equal to the component
        in another vector
        :param vector: the vector to compare against
        :return: true if every component in the vector is the same as the
        current vector, false if any component is different
        :rtype: bool
        """
        return vector.x_component == self.x_component and \
               vector.y_component == self.y_component and \
               vector.z_component == self.z_component

    def __ne__(self, vector):
        return not vector == self

    def nonzero(self):
        """ Check to see if any component of the vector is non-zero

        :rtype: bool
        :return: True if any component of the vector is nonzero, False
        otherwise
        """
        return self.x_component or self.y_component or self.z_component

    def __invert__(self):
        return Vector([-self.x_component, -self.y_component, -self.z_component])

    def mag(self):
        """ Calculate the vector's magnitude

        :return: the magnitude
        :rtype: float
        """
        return sqrt(self.x_component * self.x_component +
                    self.y_component * self.y_component +
                    self.z_component * self.z_component)

    def norm(self):
        """ Generate a vector of unit length from this vector

        :rtype: pyglet_helper.util.Vector
        :return: unit vector
        """
        magnitude = self.mag()
        if magnitude:
            # This step ensures that vector(0,0,0).norm() returns vector(0,0,0)
            # instead of NaN
            magnitude = 1.0 / magnitude
        return Vector([self.x_component * magnitude,
                       self.y_component * magnitude,
                       self.z_component * magnitude])

    def set_mag(self, magnitude):
        """
        Re-scale the vector to a set magnitude
        :param magnitude: the magnitude of the resulting vector
        :type magnitude: float
        :return:
        """
        result = self.norm() * magnitude
        self.x_component = result.x_component
        self.y_component = result.y_component
        self.z_component = result.z_component

    def __repr__(self):
        """
        Format the vector into a string
        :return:
        """
        return "Vector(" + str(self.x_component) + "," + str(self.y_component) \
               + "," + str(self.z_component) + ")"

    def dot(self, vector):
        """ Calculates the dot product of this vector and another.

        :param vector: The other vector to be multiplied with the current vector
        :type vector: Vertex or Vector
        :return: The dot product
        :rtype: float
        """
        return vector.x_component * self.x_component + \
               vector.y_component * self.y_component + \
               vector.z_component * self.z_component

    def cross(self, vector):
        """ Return the cross product of this vector and another.

        :param vector: The other vector to be crossed with the current vector
        :type vector: Vertex or Vector
        :return: The cross product of self and v
        :rtype: pyglet_helper.util.Vector
        """
        x_component = self.y_component * vector.z_component - \
                      self.z_component * vector.y_component
        y_component = self.z_component * vector.x_component \
                      - self.x_component * vector.z_component
        z_component = self.x_component * vector.y_component \
                      - self.y_component * vector.x_component
        return Vector([x_component, y_component, z_component])

    def comp(self, vector):
        """ Scalar projection of this to v

        :param vector: the vector or vertex the current vector will be compared
        to
        :type vector: Vector or Vertex
        :rtype: float
        :return: the scaling factor between v and the current vector
        """
        return self.dot(vector) / vector.mag()

    def proj(self, vector):
        """ Vector projection of this to vector

        :param vector: the vector to project the current vector onto
        :type vector: Vector
        :return: the projected vector
        :rtype: Vector
        """
        return self.dot(vector) / vector.mag() ** 2.0 * vector

    def diff_angle(self, vector):
        """ Calculates the angular difference between two vectors, in radians,
        between 0 and pi.

        :param vector: The vector to be calculated against
        :type vector: Vector
        :return: The angular difference between the current vector and vector,
        in radians
        :rtype: float
        """
        vn1 = self.norm()
        vn2 = vector.norm()
        dot_product = vn1.dot(vn2)
        if dot_product > 0.999:
            dot_product = Vector([vn2.x_component - vn1.x_component,
                                  vn2.y_component - vn1.y_component,
                                  vn2.z_component - vn1.z_component]).mag()
            return 2.0 * asin(dot_product / 2.0)
        elif dot_product < -0.999:
            dot_product = Vector([vn2.x_component + vn1.x_component,
                                  vn2.y_component + vn1.y_component,
                                  vn2.z_component + vn1.z_component]).mag()
            return pi - 2.0 * asin(dot_product / 2.0)

        return acos(dot_product)

    def scale(self, vector):
        """ Scale this vector to another, by elementwise multiplication

        :param vector: the Vector to scale by
        :type vector: Vector or Vertex
        :return: the scaled vector
        :rtype: pyglet_helper.util.Vector
        """
        return Vector([self.x_component * vector.x_component,
                       self.y_component * vector.y_component,
                       self.z_component * vector.z_component])

    def rotate(self, angle, axis=None):
        """ Rotates the vector.

        :param angle: The angle to rotate by
        :type angle: float
        :param axis: The axis to rotate around (optional, if not set, the axis
        will be the +z axis
        :type axis: Vector
        :return: rotated vector
        :rtype: pyglet_helper.util.Vector
        """
        if axis is not None:
            _rotation_matrix = rotation(angle, axis.norm())
        else:
            axis = Vector([0, 0, 1])
            _rotation_matrix = rotation(angle, axis.norm())
        return _rotation_matrix * self

    @property
    def x_component(self):
        """ Get the x component of the vector
        :return: the x component of the vector
        """
        return self._x_component

    @x_component.setter
    def x_component(self, new_value):
        """
        Set the x component of the vector
        :param new_value: the new value of the x component
        :type new_value: float
        :return:
        """
        self._x_component = new_value

    @property
    def y_component(self):
        """ Get the y component of the vector
        :return: the y component of the vector
        """
        return self._y_component

    @y_component.setter
    def y_component(self, new_value):
        """
        Set the y component of the vector
        :param new_value: the new value of the y component
        :type new_value: float
        :return:
        """
        self._y_component = new_value

    @property
    def z_component(self):
        """ Get the z component of the vector
        :return: the z component of the vector
        """
        return self._z_component

    @z_component.setter
    def z_component(self, new_value):
        """
        Set the z component of the vector
        :param new_value: the new value of the z component
        :type new_value: float
        :return:
        """
        self._z_component = new_value

    def clear(self):
        """
        Zero the state of the vector.
        :return:
        """
        self.x_component = 0.0
        self.y_component = 0.0
        self.z_component = 0.0

    def __getitem__(self, i):
        """
        get the component of the vector by an integer address
        :param i: the integer, with a value from -3 to 2
        :type i: int
        :return:
        """
        if i == 0 or i == -3:
            return self.x_component
        if i == 1 or i == -2:
            return self.y_component
        if i == 2 or i == -1:
            return self.z_component
        if i > 2 or i < -3:
            raise IndexError("index not available")

    def __setitem__(self, i, value):
        """
        set the component of the vector by an integer address
        :param i: the integer, with a value from -3 to 2
        :type i: int
        :param value: the new value to set the component to
        :type value: float
        :return:
        """
        if i == 0 or i == -3:
            self.x_component = value
        if i == 1 or i == -2:
            self.y_component = value
        if i == 2 or i == -1:
            self.z_component = value
        if i > 2 or i < -3:
            raise IndexError("index not available")

    def fabs(self):
        """
        Project the current vector into the positive quadrant
        :return: The vector in the positive quadrant
        :rtype: Vector
        """
        return Vector([abs(self.x_component), abs(self.y_component),
                       abs(self.z_component)])

    def gl_render(self):
        """
        Add the current vector as an OpenGL Vertex
        :return:
        """
        gl.glVertex3d(gl.GLdouble(self.x_component),
                             gl.GLdouble(self.y_component),
                             gl.GLdouble(self.z_component))

    def gl_normal(self):
        """
        Add the current vector as an OpenGL Normal
        :return:
        """
        gl.glNormal3dv(self.x_component)

    def sum(self):
        """
        Sum the components of the vector
        :return:
        """
        return self.x_component + self.y_component + self.z_component


class Vertex(object):
    """
     A homogeneous Vertex
    """

    def __init__(self, in_vertex=None):
        """

        :param in_vertex: a source of numeric values for the vertex, either a
        vector, vertex, list or tuple. If a vector is provided, the w component
        is assumed to be 1.
        :type in_vertex: pyglet_helper.util.Vertex, pyglet_helper.util.Vector,
        tuple, or list
        """
        self._x_component = None
        self._y_component = None
        self._z_component = None
        self._w_component = None
        if in_vertex is None:
            self.x_component = 0.0
            self.y_component = 0.0
            self.z_component = 0.0
            self.w_component = 1.0
        elif type(in_vertex) == Vector:
            self.x_component = getattr(in_vertex, 'x_component')
            self.y_component = getattr(in_vertex, 'y_component')
            self.z_component = getattr(in_vertex, 'z_component')
            self._w_component = 1.0
        elif type(in_vertex) == Vertex:
            self.x_component = getattr(in_vertex, 'x_component')
            self.y_component = getattr(in_vertex, 'y_component')
            self.z_component = getattr(in_vertex, 'z_component')
            self.w_component = getattr(in_vertex, 'w_component')
        elif type(in_vertex) is tuple or type(in_vertex) is list:
            self.x_component = in_vertex[0]
            self.y_component = in_vertex[1]
            self.z_component = in_vertex[2]
            if len(in_vertex) > 3:
                self.w_component = in_vertex[3]

    def project(self):
        """ Project the vector according to its normalization factor

        :return: A copy of the current vector scaled to w.
        :rtype: pyglet_helper.util.Vector
        """
        w_i = 1.0 / self.w_component
        return Vector([self.x_component * w_i, self.y_component * w_i,
                       self.z_component * w_i])

    def gl_render(self):
        """
        Send the vertex to OpenGl
        """
        gl.glVertex4d(self.x_component, self.y_component,
                             self.z_component, self.w_component)

    @property
    def x_component(self):
        """ Get the x component of the vertex
        :return: the x component of the vertex
        """
        return self._x_component

    @x_component.setter
    def x_component(self, new_value):
        """
        Set the x component of the vertex
        :param new_value: the new value of the x component
        :type new_value: float
        :return:
        """
        self._x_component = new_value

    @property
    def y_component(self):
        """ Get the y component of the vertex
        :return: the y component of the vertex
        """
        return self._y_component

    @y_component.setter
    def y_component(self, new_value):
        """
        Set the y component of the vertex
        :param new_value: the new value of the y component
        :type new_value: float
        :return:
        """
        self._y_component = new_value

    @property
    def z_component(self):
        """ Get the z component of the vertex
        :return: the z component of the vertex
        """
        return self._z_component

    @z_component.setter
    def z_component(self, new_value):
        """
        Set the z component of the vertex
        :param new_value: the new value of the z component
        :type new_value: float
        :return:
        """
        self._z_component = new_value

    @property
    def w_component(self):
        """ Get the w component of the vertex
        :return: the w component of the vertex
        """
        return self._w_component

    @w_component.setter
    def w_component(self, new_value):
        """
        Set the w component of the vertex
        :param new_value: the new value of the w component
        :type new_value: float
        :return:
        """
        self._w_component = new_value

    def __getitem__(self, i):
        """
        get the component of the vertex by an integer address
        :param i: the integer, with a value from -4 to 3
        :type i: int
        :return:
        """
        if i == 0 or i == -4:
            return self.x_component
        if i == 1 or i == -3:
            return self.y_component
        if i == 2 or i == -2:
            return self.z_component
        if i == 3 or i == -1:
            return self.w_component
        if i > 3 or i < -4:
            raise IndexError("index not available")

    def __setitem__(self, i, value):
        """
        set the component of the vertex by an integer address
        :param i: the integer, with a value from -4 to 3
        :type i: int
        :param value: the new value to set the component to
        :type value: float
        :return:
        """
        if i == 0 or i == -4:
            self.x_component = value
        if i == 1 or i == -3:
            self.y_component = value
        if i == 2 or i == -2:
            self.z_component = value
        if i == 3 or i == -1:
            self.w_component = value
        if i > 3 or i < -4:
            raise IndexError("index not available")


    def __repr__(self):
        """
        Format the vertex into a string
        :return:
        """
        return "Vertex(" + str(self.x_component) + "," + \
               str(self.y_component) + "," + str(self.z_component) + "," + \
               str(self.w_component) + ")"



class Tmatrix(object):
    """
    A 3D affine transformation matrix.
    """

    def __init__(self, in_tmatrix=None):
        """
        :param t: A tmatrix to copy to the current matrix
        :type t: pyglet_helper.util.Tmatrix
        """
        # This is a -precision matrix in _COLUMN MAJOR ORDER_.  User's beware.
        # It is in this order since that is what OpenGL uses internally - thus
        # eliminating a reformatting penalty.
        if in_tmatrix is not None:
            self.matrix = in_tmatrix.matrix
        else:
            self.matrix = matrix(identity(4))

    def __getitem__(self, key):
        """
        Get a specified element from a key
        :param key: the x, y component of the matrix to get
        :type key: tuple
        :return:
        """
        return self.matrix[key]

    def __mul__(self, input_data):
        """
        Multiply the data in the input by the current matrix
        :param input_data: the data to be multiplied
        :type input_data: Vector, Tmatrix, Vertex, float
        :return: self.matrix*input_data
        """
        if type(input_data) == Vector:
            out_vect = self.project(input_data)
            return out_vect
        elif type(input_data) == Tmatrix:
            tmp = Tmatrix()
            tmp.matrix = self.matrix * input_data.matrix
            return tmp
        elif type(input_data) == Vertex:
            out_vect = self.project(input_data)
            return out_vect
        else:
            tmp = Tmatrix()
            tmp.matrix = self.matrix * input_data
            return tmp

    def __rmul__(self, input_data):
        """
        Multiply the data in the input by the current matrix
        :param input_data: the data to be multiplied
        :type input_data: float or int
        :return: self.matrix*input_data
        """
        tmp = Tmatrix()
        tmp.matrix = self.matrix * input_data
        return tmp

    def inverse(self):
        """
        invert the current matrix
        :return:
        """
        self.matrix = inv(self.matrix)

    def project(self, vector):
        """ Multply a vector or vertex by the current matrix to produce a new
        vertex

        :param vector: the vertex or vector to be transformed
        :type vector: Vertex or Vector
        :return: The transformed vertex
        :rtype: pyglet_helper.util.Vertex
        """
        output_vertex = Vertex()
        if type(vector) == Vertex:
            w_component = vector.w_component
        else:
            w_component = 1.0
        output_vertex.x_component = self.matrix[0, 0] * vector.x_component + \
                        self.matrix[1, 0] * vector.y_component + \
                        self.matrix[2, 0] * vector.z_component + \
                        self.matrix[3, 0] * w_component
        output_vertex.y_component = self.matrix[0, 1] * vector.x_component + \
                        self.matrix[1, 1] * vector.y_component + \
                        self.matrix[2, 1] * vector.z_component + \
                        self.matrix[3, 1] * w_component
        output_vertex.z_component = self.matrix[0, 2] * vector.x_component + \
                        self.matrix[1, 2] * vector.y_component + \
                        self.matrix[2, 2] * vector.z_component + \
                        self.matrix[3, 2] * w_component
        output_vertex.w_component = self.matrix[0, 3] * vector.x_component + \
              self.matrix[1, 3] * vector.y_component + \
              self.matrix[2, 3] * vector.z_component + \
              self.matrix[3, 3] * w_component
        return output_vertex

    def scale(self, vector, w_component=None):
        """ Scale the transformation matrix by a vector or vertex

        :param vector: The vector or vertex describing the scaling
        :type vector: Vertex or Vector
        :param w_component: The scaling factor for the normalization column. If
        undefined, the value will be taken from the vextex, or set to 1.0
        :type w_component: float
        """
        if w_component is None:
            if type(vector) == Vertex:
                w_component = vector.w_component
            else:
                w_component = 1.0

        self.matrix[0, 0] *= vector.x_component
        self.matrix[0, 1] *= vector.x_component
        self.matrix[0, 2] *= vector.x_component
        self.matrix[0, 3] *= vector.x_component

        self.matrix[1, 0] *= vector.y_component
        self.matrix[1, 1] *= vector.y_component
        self.matrix[1, 2] *= vector.y_component
        self.matrix[1, 3] *= vector.y_component

        self.matrix[2, 0] *= vector.z_component
        self.matrix[2, 1] *= vector.z_component
        self.matrix[2, 2] *= vector.z_component
        self.matrix[2, 3] *= vector.z_component

        self.matrix[3, 0] *= w_component
        self.matrix[3, 1] *= w_component
        self.matrix[3, 2] *= w_component
        self.matrix[3, 3] *= w_component

    def translate(self, vector):
        """ Translate the transformation by a 3D value described in v

        :param vector: The vertex or vector describing the transformation
        :type vector: Vertex or Vector
        """
        self.matrix[3, 0] += vector.x_component * self.matrix[0, 0] + \
                             vector.y_component * self.matrix[1, 0] + \
                             vector.z_component * self.matrix[2, 0]
        self.matrix[3, 1] += vector.x_component * self.matrix[0, 1] + \
                             vector.y_component * self.matrix[1, 1] + \
                             vector.z_component * self.matrix[2, 1]
        self.matrix[3, 2] += vector.x_component * self.matrix[0, 2] + \
                             vector.y_component * self.matrix[1, 2] + \
                             vector.z_component * self.matrix[2, 2]
        self.matrix[3, 3] += vector.x_component * self.matrix[0, 3] + \
                             vector.y_component * self.matrix[1, 3] + \
                             vector.z_component * self.matrix[2, 3]

    def times_inv(self, vector, w_component=None):
        """ Multiply a vector or vertex v by the inverted transformation matrix

        :param vector: the vector to be transformed
        :type vector: Vector or Vertex
        :param w_component: The vector's normalization factor. If None, it will
        be taken from v or set to 1.0
        :type w_component; float
        :rtype: Vector or Vertex
        :return: The transformed Vector or Vertex
        """
        if w_component is None:
            if type(vector) == Vertex:
                w_component = vector.w_component
            else:
                w_component = 1.0
        x_component = vector.x_component - self.matrix[3, 0] * w_component
        y_component = vector.y_component - self.matrix[3, 1] * w_component
        z_component = vector.z_component - self.matrix[3, 2] * w_component
        if type(vector) == Vector:
            return Vector([self.matrix[0, 0] * x_component +
                           self.matrix[0, 1] * y_component +
                           self.matrix[0, 2] * z_component,
                           self.matrix[1, 0] * x_component +
                           self.matrix[1, 1] * y_component +
                           self.matrix[1, 2] * z_component,
                           self.matrix[2, 0] * x_component +
                           self.matrix[2, 1] * y_component +
                           self.matrix[2, 2] * z_component])
        if type(vector) == Vertex:
            return Vertex([self.matrix[0, 0] * x_component +
                           self.matrix[0, 1] * y_component +
                           self.matrix[0, 2] * z_component,
                           self.matrix[1, 0] * x_component +
                           self.matrix[1, 1] * y_component +
                           self.matrix[1, 2] * z_component,
                           self.matrix[2, 0] * x_component +
                           self.matrix[2, 1] * y_component +
                           self.matrix[2, 2] * z_component, w_component])

    def times_v(self, vector):
        """ Multiply a vector or vertex by the transformation matrix

        :param vector: the vector or vertex to be transformed
        :type vector: Vector or Vertex
        :return: the transformed vector or vertex
        :rtype: Vector or Vertex
        """
        if type(vector) == Vector:
            return Vector([self.matrix[0, 0] * vector.x_component +
                           self.matrix[1, 0] * vector.y_component +
                           self.matrix[2, 0] * vector.z_component,
                           self.matrix[0, 1] * vector.x_component +
                           self.matrix[1, 1] * vector.y_component +
                           self.matrix[2, 1] * vector.z_component,
                           self.matrix[0, 2] * vector.x_component +
                           self.matrix[1, 2] * vector.y_component +
                           self.matrix[2, 2] * vector.z_component])
        if type(vector) == Vertex:
            return Vertex([self.matrix[0, 0] * vector.x_component +
                           self.matrix[1, 0] * vector.y_component +
                           self.matrix[2, 0] * vector.z_component,
                           self.matrix[0, 1] * vector.x_component +
                           self.matrix[1, 1] * vector.y_component +
                           self.matrix[2, 1] * vector.z_component,
                           self.matrix[0, 2] * vector.x_component +
                           self.matrix[1, 2] * vector.y_component +
                           self.matrix[2, 2] * vector.z_component,
                           vector.w_component])

    def x_column(self, vector=None):
        """ Sets the first column of the matrix

        :param vector: The vector to set the first column to
        :type vector: pyglet_helper.util.Vector, pyglet_helper.util.Vertex,
        tuple or list
        """
        if vector is not None:
            if type(vector) is not Vector and type(vector) is not Vertex:
                vector = Vector(vector)
            self.matrix[0, 0] = vector.x_component
            self.matrix[0, 1] = vector.y_component
            self.matrix[0, 2] = vector.z_component

    def y_column(self, vector=None):
        """ Sets the second column of the matrix

        :param vector: The vector to set the second column to
        :type vector: pyglet_helper.util.Vector, pyglet_helper.util.Vertex,
        tuple or list
        """
        if vector is not None:
            if type(vector) is not Vector and type(vector) is not Vertex:
                vector = Vector(vector)
            self.matrix[1, 0] = vector.x_component
            self.matrix[1, 1] = vector.y_component
            self.matrix[1, 2] = vector.z_component

    def z_column(self, vector=None):
        """ Sets the third column of the matrix

        :param vector: The vector to set the third column to
        :type vector: pyglet_helper.util.Vector, pyglet_helper.util.Vertex,
        tuple or list
        """
        if vector is not None:
            if type(vector) is not Vector and type(vector) is not Vertex:
                vector = Vector(vector)

            self.matrix[2, 0] = vector.x_component
            self.matrix[2, 1] = vector.y_component
            self.matrix[2, 2] = vector.z_component

    def w_column(self, vector=None):
        """ Sets the fourth column of the matrix

        :param vector: The vector to set the fourth column to
        :type vector: pyglet_helper.util.Vector, pyglet_helper.util.Vertex,
        tuple or list
        """
        if vector is not None:
            self.matrix[3, 0] = vector.x_component
            self.matrix[3, 1] = vector.y_component
            self.matrix[3, 2] = vector.z_component
        else:
            self.matrix[3, 0] = 1.0
            self.matrix[3, 1] = 1.0
            self.matrix[3, 2] = 1.0

    def w_row(self, x_component=0, y_component=0, z_component=0, w_component=1):
        """ Set the bottom row of the matrix

        :param x_component: the value to set the first column to
        :type x_component: float
        :param y_component: the value to set the second column to
        :type y_component: float
        :param z_component: the value to set the third column to
        :type z_component: float
        :param w_component: the value to set the fourth column to
        :type w_component: float
        """
        self.matrix[0, 3] = x_component
        self.matrix[1, 3] = y_component
        self.matrix[2, 3] = z_component
        self.matrix[3, 3] = w_component

    def origin(self):
        """ Get the matrix's origin as a vector

        :return: the origin
        :rtype: pyglet_helper.util.Vector
        """
        return Vector([self.matrix[3, 0], self.matrix[3, 1], self.matrix[3, 2]])

    def gl_load(self):
        """
        Overwrites the currently active matrix in OpenGL with this one.
        """
        ctypes_matrix = (gl.GLdouble * 16)(*[float(value) for value in
                                          nditer(self.matrix)])
        gl.glLoadMatrixd(ctypes_matrix)

    def gl_mult(self):
        """
        Multiplies the active OpenGL by this one.
        """
        ctype_matrix = (gl.GLdouble * 16)(*[float(value) for value in
                                         nditer(self.matrix)])
        gl.glMultMatrixd(ctype_matrix)

    def gl_modelview_get(self):
        """ Initialize the matrix with the contents of the OpenGL modelview

        :return: the current matrix
        :rtype: matrix
        """
        ctypes_matrix = (gl.GLfloat * 16)()

        gl.glGetFloatv(gl.GL_MODELVIEW_MATRIX, ctypes_matrix)
        for i in range(0, 4):
            for j in range(0, 4):
                self.matrix[i, j] = ctypes_matrix[i + 4 * j]
        return self.matrix

    def gl_texture_get(self):
        """Initialize the matrix with the contents of the OpenGL texture matrix

        :return: the current matrix
        :rtype: matrix
        """
        ctypes_matrix = (gl.GLfloat * 16)()
        gl.glGetFloatv(gl.GL_TEXTURE_MATRIX, ctypes_matrix)
        for i in range(0, 4):
            for j in range(0, 4):
                self.matrix[i, j] = ctypes_matrix[i + 4 * j]
        return self.matrix

    def gl_color_get(self):
        """ Initialize the matrix with the contents of the OpenGL color matrix

        :return: the current matrix
        :rtype: matrix
        """
        ctypes_matrix = (gl.GLfloat * 16)()
        gl.glGetFloatv(gl.GL_COLOR_MATRIX, ctypes_matrix)
        for i in range(0, 4):
            for j in range(0, 4):
                self.matrix[i, j] = ctypes_matrix[i + 4 * j]
        return self.matrix

    def gl_projection_get(self):
        """Initialize the matrix with the contents of the OpenGL projection
        matrix

        :return: the current matrix
        :rtype: matrix
        """
        ctypes_matrix = (gl.GLfloat * 16)()
        gl.glGetFloatv(gl.GL_PROJECTION_MATRIX, ctypes_matrix)
        for i in range(0, 4):
            for j in range(0, 4):
                self.matrix[i, j] = ctypes_matrix[i + 4 * j]
        return self

    def __str__(self):
        """
        format the matrix as a string
        :return: the formatted string, four lines long
        :rtype: str
        """
        output = "| " + str(self.matrix[0, 0]) + " " +str(self.matrix[1, 0]) \
                 + " " + str(self.matrix[2, 0]) + " " + str(self.matrix[3, 0])\
                 + "|\n"
        output += "| " + str(self.matrix[0, 1]) + " " + str(self.matrix[1, 1])\
                  + " " + str(self.matrix[2, 1]) + " " + \
                  str(self.matrix[3, 1]) + "|\n"
        output += "| " + str(self.matrix[0, 2]) + " " + str(self.matrix[1, 2])\
                  + " " + str(self.matrix[2, 2]) + " " + \
                  str(self.matrix[3, 2]) + "|\n"
        output += "| " + str(self.matrix[0, 3]) + " " + str(self.matrix[1, 3])\
                  + " " + str(self.matrix[2, 3]) + " " + \
                  str(self.matrix[3, 3]) + "|\n"
        return output


def rotation(angle, axis, origin=None):
    """ Returns a rotation matrix to perform rotations about an axis passing
    through the origin through an angle in the direction specified by the Right
    Hand Rule.

    :param angle: the angle of the rotation about the axis
    :type angle: float
    :param axis: the axis of rotation
    :type axis: Vector or Vertex
    :param origin: the location of the rotation axis
    :type origin: Vector or Vertex
    :return: the new transformation matrix
    :rtype: pyglet_helper.util.Tmatrix
    """

    from math import cos, sin

    ret = Tmatrix()
    if origin is not None:
        origin = Vector(origin)
        ret = rotation(angle, axis.norm())
        rot_vect = ret * origin

        vect = origin - rot_vect
        ret.w_column(vector=vect)
    else:
        _cos = cos(angle)
        _sin = sin(angle)
        inverse_cos = 1.0 - _cos
        components = [[inverse_cos * axis[i] * axis[j] for i in range(0, 3)]
                      for j in range(0, 3)]

        ret.x_column([components[0][0] + _cos,
                      components[0][1] + axis.z_component * _sin,
                      components[0][2] - axis.y_component * _sin])
        ret.y_column([components[1][0] - axis.z_component * _sin,
                      components[1][1] + _cos,
                      components[1][2] + axis.x_component * _sin])
        ret.z_column([components[0][1] + axis.y_component * _sin,
                      components[1][2] - axis.x_component * _sin,
                      components[2][2] + _cos])

        ret.w_column()
        ret.w_row()
    return ret



"""pyglet_helper.common contains pyglet_helper libraries common to all
pyglet_helper.objects objects
"""
from pyglet_helper.common.materials import RawTexture, ShaderMaterial, \
    convert_data, load_tga, shader, TX_TURB3, TX_WOOD, TX_BRICK, TX_RANDOM, \
    LIBRARY



""" pyglet_helper.sphere contains an object for drawing a sphere
"""
try:
    import pyglet.gl as gl
except Exception as error_msg:
    gl = None
from pyglet_helper.objects import Axial, Material
from pyglet_helper.util import Quadric, Rgb, Tmatrix, Vector


class Sphere(Axial):
    """
    A Sphere object
    """
    def __init__(self, color=Rgb(), pos=Vector([0, 0, 0]),
                 axis=Vector([1.0, 0.0, 0.0]), radius=1.0,
                 material=Material(), other=None):
        """
        :param radius: The sphere's radius.
        :type radius: float
        :param color: The object's color.
        :type color: pyglet_helper.util.Rgb
        :param pos: The object's position.
        :type pos: pyglet_helper.util.Vector
        :param axis: The cone points from the base to the point along the axis.
        :type axis: pyglet_helper.util.Vector
        :param material: The object's material
        :type material: pyglet_helper.util.Material
        :param other: another sphere object to copy properties from (optional)
        :type other: pyglet_helper.objects.Sphere
        """
        super(Sphere, self).__init__(color=color, pos=pos, radius=radius,
                                     material=material, axis=axis)
        # Construct a unit sphere at the origin.
        if other is not None:
            self.axial = other
        self.compiled = False

    @property
    def scale(self):
        """ Exposed for the benefit of the ellipsoid object, which overrides it.
         The default is to use <radius, radius, radius> for the scale.
        """
        return Vector([self.radius, self.radius, self.radius])

    @property
    def material_matrix(self):
        """
        Creates a transformation matrix for spherical objects
        :return: the transformation matrix
        :rtype: pyglet_helper.util.Tmatrix
        """
        out = Tmatrix()
        out.translate(Vector([.5, .5, .5]))
        scale = self.scale
        out.scale(scale * (.5 / max(scale.x_component, max(scale.y_component,
                                                           scale.z_component))))
        return out

    @property
    def degenerate(self):
        """ Returns true if this object should not be drawn.  Conditions are:
        zero radius, or visible is false.  (overridden by the ellipsoid class).
        """
        return self.radius == 0.0

    def init_model(self, scene):
        """ Add the sphere quadrics to the View

        :param scene: The view to render the model to.
        :type scene: pyglet_helper.objects.View
        """
        sph = Quadric()

        scene.sphere_model[0].gl_compile_begin()
        sph.render_sphere(1.0, 13, 7)
        scene.sphere_model[0].gl_compile_end()

        scene.sphere_model[1].gl_compile_begin()
        sph.render_sphere(1.0, 19, 11)
        scene.sphere_model[1].gl_compile_end()

        scene.sphere_model[2].gl_compile_begin()
        sph.render_sphere(1.0, 35, 19)
        scene.sphere_model[2].gl_compile_end()

        scene.sphere_model[3].gl_compile_begin()
        sph.render_sphere(1.0, 55, 29)
        scene.sphere_model[3].gl_compile_end()

        scene.sphere_model[4].gl_compile_begin()
        sph.render_sphere(1.0, 70, 34)
        scene.sphere_model[4].gl_compile_end()

        # Only for the very largest bodies.
        scene.sphere_model[5].gl_compile_begin()
        sph.render_sphere(1.0, 140, 69)
        scene.sphere_model[5].gl_compile_end()

    def render(self, geometry):
        """ Add the sphere to the view.

        :param scene: The view to render the model into
        :type scene: pyglet_helper.objects.View
        """
        # Renders a simple sphere with the #2 level of detail.
        if self.radius == 0.0:
            return
        self.init_model(geometry)

        coverage_levels = [30, 100, 500, 5000]
        lod = self.lod_adjust(geometry, coverage_levels, self.pos, self.radius)
        gl.glPushMatrix()

        self.model_world_transform(geometry.gcf, self.scale).gl_mult()
        self.color.gl_set(self.opacity)

        if self.translucent:
            # Spheres are convex, so we don't need to sort
            gl.glEnable(gl.GL_CULL_FACE)

            # Render the back half (inside)
            gl.glCullFace(gl.GL_FRONT)
            geometry.sphere_model[lod].gl_render()

            # Render the front half (outside)
            gl.glCullFace(gl.GL_BACK)
            geometry.sphere_model[lod].gl_render()
        else:
            # Render a simple sphere.
            geometry.sphere_model[lod].gl_render()
        gl.glPopMatrix()



"""pyglet_helper.ring contains an object for drawing a ring"""
try:
    import pyglet.gl as gl
except Exception as error_msg:
    gl = None
from pyglet_helper.objects import Axial
from pyglet_helper.util import Rgb, Tmatrix, Vector
from math import pi, sin, cos, sqrt


class Ring(Axial):
    """
    A Ring object
    """
    def __init__(self, thickness=0.0, radius=1.0, color=Rgb(),
                 pos=Vector([0, 0, 0]), axis=Vector([1, 0, 0])):
        """

        :param thickness: The ring's thickness.
        :type thickness: float
        :param radius: The ring's radius.
        :type radius: float
        :param color: The object's color.
        :type color: pyglet_helper.util.Rgb
        :param pos: The object's position.
        :type pos: pyglet_helper.util.Vector
        :param axis: The cone points from the base to the point along the axis.
        :type axis: pyglet_helper.util.Vector
        """
        super(Ring, self).__init__(radius=radius, color=color, pos=pos,
                                   axis=axis)
        self._thickness = None
        self.list = None
        self.axis = axis
        # The radius of the ring's body.  If not specified, it is set to 1/10
        # of the radius of the body.
        self.thickness = thickness

    @property
    def thickness(self):
        """
        Get the ring's thickness (minor radius)
        :return: the ring's minor radius
        :rtype: float
        """
        return self._thickness

    @thickness.setter
    def thickness(self, new_thickness):
        """
        Set the ring's thickness. Will not update until render() is called
        again
        :param new_thickness: the new thickness (minor radius)
        :type new_thickness: float
        """
        self._thickness = new_thickness

    @property
    def material_matrix(self):
        """
        Creates a transformation matrix scaled to the size of the torus
        :return: the transformation matrix
        :return: pyglet_helper.util.Tmatrix
        """
        out = Tmatrix()
        out.translate(Vector([.5, .5, .5]))
        out.scale(Vector([self.radius, self.radius, self.radius]) *
                  (.5 / (self.radius + self.thickness)))
        return out

    @property
    def degenerate(self):
        """
        True if the Ring's major radius is 0
        :return:
        """
        return self.radius == 0.0

    def render(self, scene):
        """ Add a ring to the view.

        :param scene: The view to render the model into
        :type scene: pyglet_helper.objects.View
        """
        if self.degenerate:
            return
        # The number of subdivisions around the hoop's radial direction.
        if self.thickness:
            band_coverage = scene.pixel_coverage(self.pos, self.thickness)
        else:
            band_coverage = scene.pixel_coverage(self.pos, self.radius * 0.1)
        if band_coverage < 0:
            band_coverage = 1000
        bands = sqrt(band_coverage * 4.0)
        bands = clamp(4, bands, 40)
        # The number of subdivisions around the hoop's tangential direction.
        ring_coverage = scene.pixel_coverage(self.pos, self.radius)
        if ring_coverage < 0:
            ring_coverage = 1000
        rings = sqrt(ring_coverage * 4.0)
        rings = clamp(4, rings, 80)
        slices = int(rings)
        inner_slices = int(bands)
        radius = self.radius
        inner_radius = self.thickness

        # Create the vertex and normal arrays.
        vertices = []
        normals = []

        outer_angle_step = 2 * pi / (slices - 1)
        inner_angle_step = 2 * pi / (inner_slices - 1)
        outer_angle = 0.
        for i in range(slices):
            cos_outer_angle = cos(outer_angle)
            sin_outer_angle = sin(outer_angle)
            inner_angle = 0.
            for j in range(inner_slices):
                cos_inner_angle = cos(inner_angle)
                sin_inner_angle = sin(inner_angle)

                diameter = (radius + inner_radius * cos_inner_angle)
                vertex_x = diameter * cos_outer_angle
                vertex_y = diameter * sin_outer_angle
                vertex_z = inner_radius * sin_inner_angle

                normal_x = cos_outer_angle * cos_inner_angle
                normal_y = sin_outer_angle * cos_inner_angle
                normal_z = sin_inner_angle

                vertices.extend([vertex_x, vertex_y, vertex_z])
                normals.extend([normal_x, normal_y, normal_z])
                inner_angle += inner_angle_step
            outer_angle += outer_angle_step

        # Create ctypes arrays of the lists
        vertices = (gl.GLfloat *len(vertices))(*vertices)
        normals = (gl.GLfloat * len(normals))(*normals)

        # Create a list of triangle indices.
        indices = []
        for i in range(slices - 1):
            for j in range(inner_slices - 1):
                pos = i * inner_slices + j
                indices.extend([pos, pos + inner_slices, pos + inner_slices +
                                1])
                indices.extend([pos, pos + inner_slices + 1, pos + 1])
        indices = (gl.GLuint * len(indices))(*indices)

        # Compile a display list
        self.list = gl.glGenLists(1)
        gl.glNewList(self.list, gl.GL_COMPILE)
        self.color.gl_set(self.opacity)

        gl.glPushClientAttrib(gl.GL_CLIENT_VERTEX_ARRAY_BIT)
        gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
        gl.glEnableClientState(gl.GL_NORMAL_ARRAY)
        self.model_world_transform(scene.gcf,
                                   Vector([self.radius, self.radius,
                                           self.radius])).gl_mult()

        gl.glVertexPointer(3, gl.GL_FLOAT, 0, vertices)
        gl.glNormalPointer(gl.GL_FLOAT, 0, normals)
        gl.glDrawElements(gl.GL_TRIANGLES, len(indices), gl.GL_UNSIGNED_INT,
                          indices)
        gl.glPopClientAttrib()

        gl.glEndList()
        gl.glCallList(self.list)


def clamp(lower, value, upper):
    """ Restrict a value to be between a lower value and upper value. Used to
    restrict the number of polygons in the ring object

    :param lower: the lowest possible value of value
    :type lower: float or int
    :param value: the value to check
    :type value: float or int
    :param upper: the largest possible value of value
    :type upper: float or int
    :rtype: float or int
    :return: the restricted value
    """
    if lower > value:
        return lower
    if upper < value:
        return upper
    return value



""" pyglet_helper.util.color defines a few standard colors
"""
from pyglet_helper.util.rgba import Rgb

BLUE = Rgb(red=0.0, green=0.0, blue=1.0)
RED = Rgb(red=1.0, green=0.0, blue=0.0)
YELLOW = Rgb(red=1.0, green=1.0, blue=0.0)
ORANGE = Rgb(red=1.0, green=0.5, blue=0.0)
BLACK = Rgb(red=0.0, green=0.0, blue=0.0)
GREEN = Rgb(red=0.0, green=1.0, blue=0.0)
CYAN = Rgb(red=0.0, green=1.0, blue=1.0)
MAGENTA = Rgb(red=1.0, green=0.0, blue=1.0)
WHITE = Rgb(red=1.0, green=1.0, blue=1.0)
PURPLE = Rgb(red=0.5, green=0.0, blue=0.5)
GRAY = Rgb(red=0.1, green=0.1, blue=0.1)



"""
pyglet_helper.renderable contains objects needed to draw all geometric shapes
"""
try:
    import pyglet.gl as gl
except ImportError:
    gl = None

from pyglet_helper.util import DisplayList, Rgb, Tmatrix, Vector
from pyglet_helper.objects import Material

class Renderable(object):
    """
    A base class for all geometric shapes and lights.
    """

    def __init__(self, color=Rgb(), mat=Material(), opacity=1.0,
                 visible=False):
        """
        :param color: The object's color.
        :type color: pyglet_helper.util.Rgb
        :param material: The object's material
        :type material: pyglet_helper.util.Material
        :param opacity: The transparency value of the object
        :type opacity: float
        :param visible: If True, the object will be rendered on the screen
        :type visible: bool
        """
        # The base color of this body.  Ignored by the variable-color 
        # composites (curve, faces, frame).
        self.color = color
        # Fully opaque is 1.0, fully transparent is 0.0:
        self.opacity = opacity
        # True if the object should be rendered on the screen.
        self.visible = visible
        self.mat = mat

    @property
    def material(self):
        """
        Get the renderable's current material
        :return: the material
        :rtype: pyglet_helper.util.Matieral
        """
        return self.mat

    @material.setter
    def material(self, new_material):
        """
        Set the renderable's current material
        :param new_material: the new material
        :type: pyglet_helper.util.Matieral
        """
        self.mat = new_material

    @property
    def translucent(self):
        """
        True if the renderable is translucent (either the opacity is less
        than 1 or the material is translucent)
        :return:
        """
        return self.opacity != 1.0 or (self.mat and self.mat.translucent)

    def lod_adjust(self, scene, coverage_levels, pos, radius):
        """
        Calculate the level of detail required when rendering a glu object
        :param scene: the scene the the object is rendered into
        :type scene: pyglet_helper.objects.View
        :param coverage_levels: list with coverage comparison levels
        :type coverage_levels: list of int
        :param pos: the object's position within the scene
        :type pos: Vector
        :param radius: the object's size within the scene
        :type radius: float
        :return: level of detail (lod)
        :rtype: int
        """
        # See sphere.render() for a description of the level of detail calc.
        coverage = scene.pixel_coverage(pos, radius)
        lod = len(coverage_levels)
        for i in range(0, len(coverage_levels)):
            if coverage < coverage_levels[i]:
                lod = i
        lod += scene.lod_adjust
        if lod < 0:
            lod = 0
        elif lod > len(coverage_levels):
            lod = len(coverage_levels)
        return lod


class View(object):
    """
    A class for handling the environment in which all objects are rendered
    """

    def __init__(self, gcf=1.0, view_width=800, view_height=600,
                 anaglyph=False, coloranaglyph=False, forward_changed=False,
                 gcf_changed=False, lod_adjust=0, tan_hfov_x=0, tan_hfov_y=0,
                 enable_shaders=True, background_color=Rgb()):
        """
        :param gcf: The global scaling factor, a coefficient applied to all 
        objects in the view
        :type gcf: float
        :param view_width: The width of the viewport in pixels.
        :type view_width: int
        :param view_height: The height of the viewport in pixels.
        :type view_height: int
        :param anaglyph: If True, the scene will be rendered in anaglyph stereo
         mode.
        :type anaglyph: bool
        :param coloranaglyph: If True, the scene will be rendered in 
        coloranaglyph stereo mode.
        :type coloranaglyph: bool
        :param forward_changed: True if the forward vector changed since the 
        last rending operation.
        :type forward_changed: bool
        :param gcf_changed: True if the global scaling factor changed since 
        the last render cycle.
        :type gcf_changed: bool
        :param lod_adjust: The level-of-detail.
        :type lod_adjust: int
        :param tan_hfov_x: The tangent of half the horizontal field of view.
        :type tan_hfov_x: int
        :param tan_hfov_y: The tangent of half the vertical field of view.
        :type tan_hfov_y: int
        :param enable_shaders: If True, shader programs will be allowed
        :type enable_shaders: bool
        :param background_color: The scene's background color
        :type background_color: pyglet_helper.util.Rgb
        """
        # The position of the camera in world space.
        self.camera = Vector()
        # The direction the camera is pointing - a unit vector.
        self.forward = Vector()
        # The center of the scene in world space.
        self.center = Vector()
        # The true up direction of the scene in world space.
        self.up_vector = Vector()
        self.view_width = view_width
        self.view_height = view_height
        self.forward_changed = forward_changed
        self.gcf = gcf
        # The vector version of the Global Scaling Factor, for scene.uniform=0
        self.gcfvec = Vector()
        self.gcf_changed = gcf_changed
        self.lod_adjust = lod_adjust
        self.anaglyph = anaglyph
        self.coloranaglyph = coloranaglyph
        self.tan_hfov_x = tan_hfov_x
        self.tan_hfov_y = tan_hfov_y

        self.box_model = DisplayList()
        self.sphere_model = [DisplayList()] * 6
        self.cylinder_model = [DisplayList()] * 6
        self.cone_model = [DisplayList()] * 6
        self.pyramid_model = DisplayList()

        self.camera_world = Tmatrix()

        self.background_color = background_color
        self.lights = []

        self.enable_shaders = enable_shaders
        self.screen_objects = []
        self.is_setup = False
        self.setup()

    def setup(self):
        """ Does some one-time OpenGL setup.
        """
        gl.glEnable(gl.GL_LIGHTING)
        gl.glClearColor(1, 1, 1, 1)
        gl.glColor3f(1, 0, 0)
        gl.glEnable(gl.GL_DEPTH_TEST)
        gl.glEnable(gl.GL_CULL_FACE)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
        gl.glLoadIdentity()
        self.is_setup = True

    def draw_lights(self):
        """ Render the lights in the scene
        """
        GL_DEFINED_LIGHTS = [gl.GL_LIGHT0, gl.GL_LIGHT1,
                         gl.GL_LIGHT2, gl.GL_LIGHT3,
                         gl.GL_LIGHT4, gl.GL_LIGHT5,
                         gl.GL_LIGHT6, gl.GL_LIGHT7]
        max_lights = min(len(self.lights), 8)
        # add all of the lights to the scene
        for i in range(0, max_lights):
            # enable all of the lights
            gl.glEnable(GL_DEFINED_LIGHTS[i])
            gl.glLightfv(GL_DEFINED_LIGHTS[i], gl.GL_POSITION, self.lights[
                i].position)
            gl.glLightfv(GL_DEFINED_LIGHTS[i], gl.GL_SPECULAR,
                      self.lights[i].specular)
            gl.glLightfv(GL_DEFINED_LIGHTS[i], gl.GL_DIFFUSE,
                                self.lights[i].diffuse)

    def pixel_coverage(self, pos, radius):
        """ Compute the apparent diameter, in pixels, of a circle that is
        parallel to the screen, with a center at pos, and some radius.  If pos
        is behind the camera, it will return negative.

        :param pos: The position in the view to examine
        :type pos: pyglet_helper.util.Vector
        :param radius: The radius of coverage to examine
        :type radius: float
        :return: The diameter in pixels.
        :rtype: float
        """
        # The distance from the camera to this position, in the direction of 
        # the camera.  This is the distance to the viewing plane that the 
        # coverage circle lies in.

        pos = Vector(pos)
        dist = (pos - self.camera).dot(self.forward)
        # Half of the width of the viewing plane at this distance.
        apparent_hwidth = self.tan_hfov_x * dist
        # The fraction of the apparent width covered by the coverage circle.
        if apparent_hwidth == 0:
            coverage_fraction = 1
        else:
            coverage_fraction = radius / apparent_hwidth
        # Convert from fraction to pixels.
        return coverage_fraction * self.view_width



# -*- coding: utf-8 -*-
#
# PygletHelper documentation build configuration file, created by
# sphinx-quickstart on Thu Dec 31 17:55:20 2015.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.


# -- General configuration ------------------------------------------------


# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'


# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'PygletHelper'
copyright = '2015, Catherine Holloway'
author = 'Catherine Holloway'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '0.0.1'
# The full version, including alpha/beta/rc tags.
release = '0.0.1'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None


# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']


# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'alabaster'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Output file base name for HTML help builder.
htmlhelp_basename = 'PygletHelperdoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [(master_doc, 'PygletHelper.tex',
                    'PygletHelper Documentation', 'Catherine Holloway',
                    'manual')]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, 'pyglethelper',
              'PygletHelper Documentation', [author], 1)]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [(master_doc, 'PygletHelper',
                      'PygletHelper Documentation', author,
                      'PygletHelper', 'One line description of project.',
                      'Miscellaneous')]

autoclass_content = 'both'



""" pyglet_helper.util.texture contains objects for describing textures to
apply to objects
"""
from __future__ import print_function
try:
    from pyglet.gl import glBindTexture, glDeleteTextures, GL_TEXTURE_2D
except Exception as error_msg:
    print("Pyglet import error: "+str(error_msg))

class Texture(object):
    """
    A class to assist in managing OpenGL texture resources.
    """
    def __init__(self, damaged=False, handle=0, opacity=False):
        self._have_opacity = None
        self.damaged = damaged
        self.handle = handle
        # A unique identifier for the texture, to be obtained from
        # glGenTextures().
        self._opacity = opacity

    @property
    def opacity(self):
        """
        Get the texture's opacity
        :return: the texture's opacity, from 0 to 1
        :rtype: float
        """
        return self._have_opacity

    @opacity.setter
    def opacity(self, opacity):
        """
        Set the texture's opacity
        :param opacity: the new opacity; a float from 0 to 1
        :type opacity: float
        """
        self._have_opacity = opacity

    def gl_activate(self):
        """
        Make this texture active.  This function constitutes use under the
            "initialize on first use" rule, and will incur a one-time speed and
            continuous graphics memory penalty.  Precondition: an OpenGL
            context must be active.
        """
        if self.damaged:
            self.damaged = False
        if not self.handle:
            return

        glBindTexture(GL_TEXTURE_2D, self.handle)

    def gl_free(self):
        """
        Returns e.g. GL_TEXTURE_2D - the thing to be enabled to make this
            texture work with the fixed function pipeline.
        """
        print("Deleting texture number " + self.handle)
        glDeleteTextures(1, self.handle)
        # Mutable subclasses must call this function whenever their texture
        # data

    def damage(self):
        """
        Damage the texture, indicating that it needs to be re-uploaded to
        OpenGL
        """
        self.damaged = True



"""
pyglet_helper.cylinder contains an object for drawing a cylinder
"""
try:
    import pyglet.gl as gl
except Exception as err_msg:
    gl = None
from pyglet_helper.objects import Axial
from pyglet_helper.util import Quadric, Rgb, Vector


class Cylinder(Axial):
    """
     A Cylinder object.
    """
    def __init__(self, pos=Vector([0, 0, 0]), axis=Vector([1, 0, 0]),
                 radius=1.0,
                 color=Rgb(0, 0, 0)):
        """

        :param radius: The cylinder's radius.
        :type radius: float
        :param color: The object's color.
        :type color: pyglet_helper.util.Rgb
        :param pos: The object's position.
        :type pos: pyglet_helper.util.Vector
        :param axis: The cone points from the base to the point along the axis.
        :type axis: pyglet_helper.util.Vector
        """
        super(Cylinder, self).__init__(pos=pos, radius=radius, color=color,
                                       axis=axis)

    def init_model(self, scene):
        """ Add the cylinder quadrics to the view.

        :param scene: The view to render the model to.
        :type scene: pyglet_helper.objects.View
        """
        # The number of faces corresponding to each level of detail.
        n_faces = [8, 16, 32, 64, 96, 188]
        n_stacks = [1, 1, 3, 6, 10, 20]
        for i in range(0, 6):
            scene.cylinder_model[i].gl_compile_begin()
            _quadric = Quadric()
            _quadric.render_cylinder(1.0, 1.0, n_faces[i], n_stacks[i])
            gl.glTranslatef(1.0, 0.0, 0.0)
            _quadric.render_disk(1.0, n_faces[i], 1, 1)  # left end of cylinder
            gl.glTranslatef(-1.0, 0.0, 0.0)
            _quadric.render_disk(1.0, n_faces[i], 1, -1)  # right end of
            scene.cylinder_model[i].gl_compile_end()

    @property
    def degenerate(self):
        """
        Returns True if the cylinder is invisible or has length or radius of 0
        :return:
        """
        return not self.visible or self.radius == 0.0 or self.axis.mag() == 0.0

    def render(self, scene):
        """ Add the cylinder to the view.

        :param scene: The view to render the model into
        :type scene: pyglet_helper.objects.View
        """
        if self.radius == 0.0:
            return
        self.init_model(scene)

        coverage_levels = [10, 25, 50, 196, 400]
        lod = self.lod_adjust(scene, coverage_levels, self.pos, self.radius)

        gl.glPushMatrix()
        self.model_world_transform(scene.gcf, Vector([self.length, self.radius,
                                                      self.radius])).gl_mult()

        if self.translucent:
            gl.glEnable(gl.GL_CULL_FACE)
            self.color.gl_set(self.opacity)

            # Render the back half.
            gl.glCullFace(gl.GL_FRONT)
            scene.cylinder_model[lod].gl_render()

            # Render the front half.
            gl.glCullFace(gl.GL_BACK)
            scene.cylinder_model[lod].gl_render()
        else:
            self.color.gl_set(self.opacity)
            scene.cylinder_model[lod].gl_render()
        gl.glPopMatrix()

    @property
    def center(self):
        """
        :return: Position + axis/2
        :rtype: float
        """
        return self.pos + self.axis * 0.5



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Python setup.py file for the pyglet_helper project
"""

# IMPORTS ####################################################################

import codecs
import os
import re

from setuptools import setup, find_packages

# SETUP VALUES ###############################################################

NAME = "pyglet_helper"
PACKAGES = find_packages()
META_PATH = os.path.join("pyglet_helper", "__init__.py")
INSTALL_REQUIRES = ['enum34', 'numpy', 'pyglet']

# HELPER FUNCTONS ############################################################

HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    with codecs.open(os.path.join(HERE, *parts), "rb", "utf-8") as f:
        return f.read()

META_FILE = read(META_PATH)


def find_meta(meta):
    """
    Extract __*meta*__ from META_FILE.
    """
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta),
        META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))


# MAIN #######################################################################

def main():
    setup(
        name=find_meta("title"),
        version=find_meta("version"),
        url=find_meta("uri"),
        author=find_meta("author"),
        author_email=find_meta("email"),
        packages=PACKAGES,
        description=find_meta("description"),
        install_requires=INSTALL_REQUIRES,
        package_data = {"pyglet_helper": ['common/library.txt']}
    )

if __name__ == "__main__":
    main()



"""
pyglet_helper.rectangular contains a base class for drawing objects with
rectangular symmetry
"""

from pyglet_helper.objects import Primitive
from pyglet_helper.util import Rgb, Vector


class Rectangular(Primitive):
    """
    A base class for rectangular-like objects (such as the box or pyramid)
    """
    def __init__(self, axis=Vector([1, 0, 0]), pos=Vector([0, 0, 0]),
                 width=1.0, height=1.0, length=1.0, color=Rgb(), other=None):
        """
        :param color: The object's color.
        :type color: pyglet_helper.util.Rgb
        :param pos: The object's position.
        :type pos: pyglet_helper.util.Vector
        :param width: The object's width.
        :type width: float
        :param height: The object's height.
        :type height: float
        :param length: The object's length.
        :type length: float
        """
        super(Rectangular, self).__init__(color=color, pos=pos, axis=axis)
        self._height = None
        self._width = None
        if other is None:
            self.width = width
            self.height = height
        else:
            self.width = other.width
            self.height = other.height
        self.length = length

    @property
    def scale(self):
        """
        Gets the rectangle's scale (defined as its length
        :return: the rectangle's scale
        :rtype: float
        """
        return self.length

    def apply_transform(self, scene):
        """ Scale the object to the correct height, width and length

        :param scene: the object's current view
        :type scene: pyglet_helper.objects.View
        """
        min_scale = max(self.axis.mag(), max(self.height, self.width)) * 1e-6
        self.size = Vector([max(min_scale, self.axis.mag()),
                            max(min_scale, self.height),
                            max(min_scale, self.width)])
        self.model_world_transform(scene.gcf, self.size).gl_mult()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Defines globally-available subpackages and symbols for the pyglet_helper
package.
"""

from __future__ import absolute_import

from . import objects
from . import common
from . import util

__version__ = "0.0.1"
__author__ = "cholloway"

__title__ = "pyglet_helper"
__description__ = "3d modelling library"
__uri__ = "http://catherineh.github.io/pyglet_helper/"


__email__ = "milankie@gmail.com"

__license__ = "AGPLv3"
__copyright__ = "Copyright (c) 2014-2016 Catherine Holloway"



""" pyglet_helper.material contains an object for describing materials which
can be applied to objects
"""
from pyglet_helper.util import ShaderProgram, Texture


class Material(object):
    """
     A Material object, used for storing and handling shader programs.
    """
    def __init__(self, translucent=False, shader_program=None):
        """

        :param translucent:
        :type translucent: bool
        :param shader_program: The GLSL (Graphics library shader language) code
         to be executed).
        :type shader_program: str
        :return:
        """

        self._textures = None
        self._shader = None
        self._translucent = None
        self.translucent = translucent
        self.textures = [Texture()]
        self.shader = shader_program

    @property
    def textures(self):
        """
        Gets the material's texture
        :return: the texture
        """
        return self._textures

    @textures.setter
    def textures(self, new_texture):
        """
        Sets the material's texture
        :param new_texture: the new texture
        """
        self._textures = new_texture

    @property
    def shader(self):
        """
        Gets the shader program, if it exists
        :return: the shader program
        :rtype: pyglet_helper.util.ShaderProgram
        """
        if hasattr(self, '_shader') and type(self._shader) == 'shader_program':
            return self._shader
        else:
            return None

    @shader.setter
    def shader(self, source):
        """
        Sets the shader source
        :param source: the source of a new shader program
        :type source: str
        """
        if source is not None:
            self._shader = ShaderProgram(source)
        else:
            self._shader = None

    @property
    def translucent(self):
        """
        Gets the translucence value of the material
        :return: the translucence value
        :rtype: float
        """
        return self._translucent

    @translucent.setter
    def translucent(self, new_translucent):
        """
        Sets the translucence value of the material
        :param new_translucent: the translucence value
        :type new_translucent: float
        """
        self._translucent = new_translucent

    def get_shader_program(self):
        """
        Gets the current ShaderProgram
        """
        if type(self.shader) == ShaderProgram:
            return self.shader.get()
        else:
            return None


UNSHADED = Material(shader_program="""
[vertex]
        void main() {
            gl_Position = ftransform();
            gl_FrontColor = gl_Color;
        }
[fragment]
        void main() {
            gl_FragColor = gl_Color;
        }""")
EMISSIVE = Material(shader_program="""
    [fragment]
    void material_main() {
        float d = dot(normalize(-position), normal);
        d = pow(d * 1.5,.4) * 1.1;
        if (d > 1.0) d = 1.0;
        material_color = object_color * d;
        material_opacity = object_opacity;
    }""")
DIFFUSE = Material(shader_program="""
        [fragment]
        void material_main() {
            material_color = lightAt( normalize(normal), normalize(-position), object_color, vec3(0,0,0), 0.0 );
            material_opacity = object_opacity;
        }""")
PLASTIC = Material(shader_program="""
        [fragment]
        void material_main() {
            material_color = lightAt( normalize(normal), normalize(-position), object_color, vec3(.8,.8,.8), 64.0 );
            material_opacity = object_opacity;
        }
        """)
ROUGH = Material(shader_program="""
        [fragment]
        uniform sampler3D tex0;

        void material_main() {
            // Displace the surface normal using a 3D noise function
            vec3 N2 = normal + noise3D(tex0, mat_pos, 1.0) * .4;

            // Compute lighting based on the displaced normal
            material_color = lightAt( normalize(N2), normalize(-position),
                                        object_color,
                                        vec3(.5,.5,.5),
                                        16. )
                                // TODO hack to reduce ambient
                                - object_color * gl_LightModel.ambient.rgb * .7;

            material_opacity = object_opacity;
        }
        """)
SHINY = Material(shader_program="""
        [fragment]
        uniform sampler3D tex0;

        void material_main() {
            // Displace the surface normal using a 3D noise function
            vec3 N2 = normal/2.0 + noise3D(tex0, acos(normal/mat_pos), 0.2) * 1.4;

            // Compute lighting based on the displaced normal
            material_color = lightAt( normalize(N2), normalize(position),
                                        object_color,
                                        vec3(1.0,1.0,1.0),
                                        16. )
                                // TODO hack to reduce ambient
                                - object_color * gl_LightModel.ambient.rgb * .7;

            material_opacity = object_opacity;
        }
        """)
CHROME = Material(shader_program="""
        [fragment]
        uniform sampler3D tex0;

        void material_main() {
            // Displace the surface normal using a 3D noise function
            vec3 N2 = normal/2.0 + noise3D(tex0, log((position/35.0+0.5)), 0.2) * 2.4;

            // Compute lighting based on the displaced normal
            material_color = lightAt( normalize(cross(N2,mat_pos)), normalize(-cross(mat_pos,normal)),
                                        object_color*((2.,2.,2.)-gl_LightModel.ambient.rgb),
                                        vec3(0.5,0.5,0.5),
                                        0.2 )
                                // TODO hack to reduce ambient
                                - object_color * gl_LightModel.ambient.rgb * .7;

            material_opacity = 1.0;
        }
        """),

ICE = Material(shader_program="""
        [fragment]
        uniform sampler3D tex0;

        void material_main() {
            // Displace the surface normal using a 3D noise function
            vec3 N2 = normal/2.0 + noise3D(tex0, log2((position/5.791+0.5)), 0.2) * 1.4;

            // Compute lighting based on the displaced normal
            material_color = lightAt( normalize(cross(N2,mat_pos)), normalize(-cross(position,normal)),
                                        vec3(1.0,0.9,1.0),
                                        vec3(0.5,0.5,0.5),
                                        0.1 )
                                // TODO hack to reduce ambient
                                - object_color * gl_LightModel.ambient.rgb * .7;

            material_opacity = 0.5;
        }
        """)

GLASS = Material(shader_program="""
        [varying]
        varying vec3 gln;
        [vertex]
        void main(void) {
            basic();
            gln = gl_Normal;
        }
        [fragment]
        uniform sampler3D tex0;

        void material_main() {
            // Displace the surface normal using a 3D noise function
            // vec3 N2 = normal/1.0 + noise3D(tex0, acos(normal/mat_pos), 0.2) * 2.4;

            material_color = lightAt( normalize(normal), normalize(-position), object_color, vec3(.7,.7,.7), 32.0 );

            material_opacity = 0.4; // dot(cross(gln,-position),mat_pos);
        }
        """)

BLAZED = Material(shader_program="""
        [fragment]
        uniform sampler3D tex0;

        void material_main() {
            // Displace the surface normal using a 3D noise function
            vec3 N2 = normal + noise3D(tex0, sqrt((normal+0.5)/(position+0.5)), 0.2) * 1.4;

            // Compute lighting based on the displaced normal
            material_color = lightAt( normalize(cross(N2,mat_pos)), normalize(-cross(position,normal)),
                                        object_color,
                                        vec3(1.0,1.0,1.0),
                                        16. )
                                // TODO hack to reduce ambient
                                - object_color * gl_LightModel.ambient.rgb * .7;

            material_opacity = 1.0;
        }
        """),
SILVER = Material(shader_program="""
        [fragment]
        uniform sampler3D tex0;

        void material_main() {
            // Displace the surface normal using a 3D noise function
            vec3 N2 = normal/2.0 + noise3D(tex0, mat_pos*32.0, 1.0) * 0.2;

            // Compute lighting based on the displaced normal
            material_color = lightAt( normalize(N2), normalize(-position),
                                        object_color,
                                        vec3(0.5,0.5,0.5),
                                        0.2 )
                                // TODO hack to reduce ambient
                                - object_color * gl_LightModel.ambient.rgb * .7;

            material_opacity = 1.0;
        }
        """)

WOOD = Material(shader_program="""
        [fragment]
        uniform sampler2D tex0;  // wood cross-section
        uniform sampler3D tex1;  // 3D turbulence

        void material_main() {
            // Compute a position in the 2D cross section texture
            vec2 wood_pos = -.85 * mat_pos.zy +    //< simple rectangular mapping
                            -.10 * mat_pos.x +     //< slight skew to display wood grain on xy and xz surfaces
                            -.05 * noise3D(tex1, mat_pos * .5, 1.).xy;   //< turbulence so grain isn't perfectly straight

            // Look up the color in the texture
            vec3 C = texture2D( tex0, wood_pos ).rgb;

            // Apply lighting
            material_color = lightAt( normalize(normal), normalize(-position),
                                        C*object_color,
                                        vec3(.5,.5,.5),
                                        5. );
            material_opacity = object_opacity;
        }
        """)
MARBLE = Material(shader_program="""
        [fragment]
        uniform sampler3D tex0;

        void material_main() {
            vec3 noise = noise3D( tex0, mat_pos, 2.0 );

            // "marble" varies between two colors in a sine wave pattern in y,
            //    displaced heavily by a noise function.
            float a = 0.5 + 0.5*sin( mat_pos.y*16. + noise.x*10. );
            vec3 C = mix( vec3(.4,.3,.3), vec3(1.,1.,1.), a );

            // We are also doing a normal displacement similar to "rough".
            // TODO: Is this desired?  I normally think of marble as smooth!
            vec3 N2 = normal + noise*1.;

            // Modulate the marble color by the object color, and apply lighting
            material_color = lightAt( normalize(N2), normalize(-position), C*object_color, vec3(.8,.8,.8), 100. );
            material_opacity = object_opacity;
        }
        """),
EARTH = Material(shader_program="""
        [fragment]
        void material_main() {
            material_color = mat_pos * .5;
            if ( fract( mat_pos.x * 10. + .021 ) < .02 ) material_color *= 0.5;
            if ( fract( mat_pos.y * 10. + .021 ) < .02 ) material_color *= 0.5;
            if ( fract( mat_pos.z * 10. + .021 ) < .02 ) material_color *= 0.5;
            if ( mat_pos.x < 0. || mat_pos.x > 1. ) material_color = vec3(1.,0.8,0.8);
            if ( mat_pos.y < 0. || mat_pos.y > 1. ) material_color = vec3(0.8,1.,0.8);
            if ( mat_pos.z < 0. || mat_pos.z > 1. ) material_color = vec3(0.8,0.8,1.);
            material_opacity = object_opacity;
        }
        """)
# fancy earth renderer with clouds
BLUEMARBLE = Material(shader_program="""
        [fragment]
        void material_main() {
            material_color = mat_pos * .5;
            if ( fract( mat_pos.x * 10. + .021 ) < .02 ) material_color *= 0.5;
            if ( fract( mat_pos.y * 10. + .021 ) < .02 ) material_color *= 0.5;
            if ( fract( mat_pos.z * 10. + .021 ) < .02 ) material_color *= 0.5;
            if ( mat_pos.x < 0. || mat_pos.x > 1. ) material_color = vec3(1.,0.8,0.8);
            if ( mat_pos.y < 0. || mat_pos.y > 1. ) material_color = vec3(0.8,1.,0.8);
            if ( mat_pos.z < 0. || mat_pos.z > 1. ) material_color = vec3(0.8,0.8,1.);
            material_opacity = object_opacity;
        }
        """)
BRICKS = Material(shader_program="""
        [fragment]
        void material_main() {
            material_color = mat_pos;
            material_opacity = object_opacity;
        }
        """)



"""
pyglet_helper.primitive contains objects and methods related to drawing all
geometric shapes
"""
from pyglet_helper.objects import Material, Renderable
from pyglet_helper.util import Rgb, rotation, Tmatrix, Vector



class Primitive(Renderable):
    """
     A base class for all geometric shapes.
    """
    def __init__(self, axis=Vector([1, 0, 0]), up_vector=Vector([0, 1, 0]),
                 pos=Vector([0, 0, 0]), obj_initialized=False, color=Rgb(),
                 material=Material()):
        """

        :param axis: The orientation to use when drawing.
        :type axis: pyglet_helper.util.Vector
        :param up_vector: A vector that points to the current up direction in
        the view.
        :type up_vector: pyglet_helper.util.Vector
        :param pos: The object's position.
        :type pos: pyglet_helper.util.Vector
        :param obj_initialized: If True, the object has been initialized
        :type obj_initialized: bool
        :param color: The object's color.
        :type color: pyglet_helper.util.Rgb
        :param material: The object's material
        :type material: pyglet_helper.util.Material
        """
        super(Primitive, self).__init__(color=color, mat=material)
        self._axis = None
        self._pos = None
        self._up = None
        self._width = None
        self._height = None

        self.startup = True
        self.obj_initialized = obj_initialized

        # position must be defined first, before the axis
        self.up_vector = Vector(up_vector)
        self.pos = Vector(pos)
        self.axis = Vector(axis)

    def model_world_transform(self, world_scale=0.0,
                              object_scale=Vector([1, 1, 1])):
        """Performs scale, rotation, translation, and world scale (gcf)
        transforms in that order.

        :param world_scale: The global scaling factor.
        :type world_scale: float
        :param object_scale: The scaling to applied to this specific object
        :type object_scale: pyglet_helper.util.Vector
        :rtype: pyglet_helper.util.Tmatrix
        :returns:  Returns a tmatrix that performs reorientation of the object
        from model orientation to world
         (and view) orientation.
        """
        ret = Tmatrix()
        # A unit vector along the z_axis.
        z_axis = Vector([0, 0, 1])
        if abs(self.axis.dot(self.up_vector) / self.up_vector.mag()**2.0) \
                > 0.98:
            # Then axis and up are in (nearly) the same direction: therefore,
            # try two other possible directions for the up vector.
            if abs(self.axis.norm().dot(Vector([-1, 0, 0]))) > 0.98:
                z_axis = self.axis.cross(Vector([0, 0, 1])).norm()
            else:
                z_axis = self.axis.cross(Vector([-1, 0, 0])).norm()
        else:
            z_axis = self.axis.cross(self.up_vector).norm()

        y_axis = z_axis.cross(self.axis).norm()
        x_axis = self.axis.norm()
        ret.x_column(x_axis)
        ret.y_column(y_axis)
        ret.z_column(z_axis)
        w_column = world_scale*self.pos
        ret.w_column(w_column)
        ret.w_row()

        ret.scale(object_scale * world_scale, 1)

        return ret

    def rotate(self, angle, axis, origin):
        """Rotate the primitive's axis by angle about a specified axis at a
        specified origin.

        :param angle: the angle to rotate by, in radians
        :type angle: float
        :param axis: The axis to rotate around.
        :type axis: pyglet_helper.util.Vector
        :param origin: The center of the axis of rotation.
        :type origin: pyglet_helper.util.Vector
        """
        rotation_matrix = rotation(angle, axis, origin)
        fake_up = self.up_vector
        if not self.axis.cross(fake_up):
            fake_up = Vector([1, 0, 0])
            if not self.axis.cross(fake_up):
                fake_up = Vector([0, 1, 0])
        # is this rotation needed at present? Is it already included in the
        # transformation matrix?
        #self.pos = R * self._pos
        self.up_vector = rotation_matrix.times_v(fake_up)
        self._axis = rotation_matrix.times_v(self._axis)

    @property
    def center(self):
        """
        Gets the object's center
        :return: the object's center, as a vector
        :rtype: pyglet_helper.util.Vector
        """
        return self.pos

    @property
    def pos(self):
        """
        Get the object's current position
        :return: the object's position
        :rtype: pyglet_helper.util.Vector
        """
        return self._pos

    @pos.setter
    def pos(self, n_pos):
        """
        Set the object's position with a Vector
        :param n_pos: a vector with the object's new position
        :type n_pos: pyglet_helper.util.Vector
        :return:
        """
        self._pos = Vector(n_pos)

    @property
    def length(self):
        """
        Get the object's length (for most objects, this is the magnitude of
        its axis)
        :return: the object's length
        :rtype: float
        """
        return self.axis.mag()

    @length.setter
    def length(self, new_length):
        """
        Set the object's length
        :param new_length: the new length of the object
        :type new_length: float
        :return:
        """
        if new_length < 0:
            raise ValueError("length cannot be negative")
        self.axis = self.axis.norm() * new_length

    @property
    def height(self):
        """
        Gets the scale of the object along the y axis
        :return: the object's scale along the y axis
        :rtype: float
        """
        return self._height

    @height.setter
    def height(self, new_height):
        """
        Sets the scale of the object along the y axis
        :param new_height: the object's scale along the y axis
        :type new_height: float
        """
        if new_height < 0:
            raise ValueError("height cannot be negative")
        self._height = new_height

    @property
    def width(self):
        """
        Gets the scale of the object along the z axis
        :return: the object's scale along the z axis
        :rtype: float
        """
        return self._width

    @width.setter
    def width(self, new_width):
        """
        Sets the scale of the object along the z axis
        :param new_width: the object's scale along the y axis
        :type new_width: float
        """
        if new_width < 0:
            raise ValueError("width cannot be negative")
        self._width = new_width

    @property
    def size(self):
        """
        Gets a vector with the object's scale along the x, y, and z axes
        :return: a vector with the scales
        :rtype: pyglet_helper.util.Vector
        """
        return Vector([self.axis.mag(), self.height, self.width])

    @size.setter
    def size(self, new_size):
        """
        Sets the object's scale along the x, y, and z axes
        :param new_size: the new scales, in a vector
        :type new_size: pyglet_helper.util.Vector
        """
        if new_size.x_component < 0:
            raise ValueError("length cannot be negative")
        if new_size.y_component < 0:
            raise ValueError("height cannot be negative")
        if new_size.z_component < 0:
            raise ValueError("width cannot be negative")
        self.axis = self.axis.norm() * new_size.x_component
        self.height = new_size.y_component
        self.width = new_size.z_component

    @property
    def axis(self):
        """
        Get the object's axis, which defines the orientation and size of the
        object
        :return: the object's axis
        :rtype; pyglet_helper.util.Vector
        """
        return self._axis

    @axis.setter
    def axis(self, n_axis):
        """
        Set the object's axis, as a vector
        :param n_axis: the new axis
        :type n_axis: pyglet_helper.util.Vector
        """
        if self.axis is None:
            self._axis = Vector([1, 0, 0])
        _axis = self.axis.cross(n_axis)
        if _axis.mag() == 0.0:
            self._axis = n_axis
        else:
            angle = n_axis.diff_angle(self._axis)
            self._axis = n_axis.mag() * self._axis.norm()
            self.rotate(angle, _axis, self.pos)

    @property
    def up_vector(self):
        """
        Get the object's up axis
        :return: the object's up axis
        :rtype: pyglet_helper.util.Vector
        """
        return self._up

    @up_vector.setter
    def up_vector(self, n_up):
        """
        Set the object's up axis
        :param n_up: the object's up axis
        :type n_up: pyglet_helper.util.Vector
        """
        self._up = n_up

    @property
    def is_light(self):
        """
        Returns false if the object is not a light. By default,
        all primitives are not lights
        :return: whether the object is a light
        :rtype: bool
        """
        return False



""" pyglet_helper.util contains objects useful to creating materials and
geometric shapes
"""
from pyglet_helper.util.color import BLUE, RED, GREEN, GRAY, MAGENTA, YELLOW, \
    CYAN, ORANGE, BLACK, PURPLE, WHITE
from pyglet_helper.util.display_list import DisplayList
from pyglet_helper.util.quadric import DrawingStyle, NormalStyle, \
    Orientation, Quadric
from pyglet_helper.util.rgba import Rgba, Rgb
from pyglet_helper.util.shader_program import ShaderProgram, UseShaderProgram
from pyglet_helper.util.texture import Texture
from pyglet_helper.util.linear import Vector, Vertex, Tmatrix, rotation




""" pyglet_helper.display_list contains a class for generating lists of
objects to render to the screen
"""
from __future__ import print_function
try:
    import pyglet.gl as gl
except ImportError:
    gl = None

class DisplayList(object):
    """
    A class for storing the OpenGl commands for rendering an object.
    """
    def __init__(self, built=False):
        """
        :param built: If True, the commands have been executed
        :type built: bool
        """
        self.built = built
        self.handle = gl.glGenLists(1)

    def gl_compile_begin(self):
        """ Generates the beginning of the list.
        """
        gl.glNewList(self.handle, gl.GL_COMPILE)

    def gl_compile_end(self):
        """ Generates the end of the list.
        """
        gl.glEndList(self.handle)
        self.built = True

    def gl_render(self):
        """ Call all of the commands in the current list.
        """
        try:
            gl.glCallList(self.handle)
        except gl.GLException as e_msg:
            print("Got GL Exception on call list: " + str(e_msg))
        self.built = True

    @property
    def compiled(self):
        """ Returns whether the current list has beein completed.
        """
        return self.built
